#!/usr/bin/env python3
'''
Addon functions that need additional dependencies to work.

To also install all dependecies to use them, use::

    pip install .[addons]
'''
from .fods import get_fods, remove_core_fods, split_atom_and_fod
from .viewer import view_grid, view_mol
from .workflows import FO, FLO, KSO

__all__ = ['FO', 'FLO', 'get_fods', 'KSO', 'remove_core_fods', 'split_atom_and_fod', 'view_grid',
           'view_mol']



#!/usr/bin/env python3
'''
Calculate different energy contributions.
'''
import numpy as np
from numpy.linalg import inv
from scipy.special import erfc

from .exc import get_exc


class Energy:
    '''Energy class to save energy contributions in one place.'''
    __slots__ = ['Ekin', 'Eloc', 'Enonloc', 'Ecoul', 'Exc', 'Eewald', 'Esic']

    def __init__(self):
        self.Ekin = 0
        self.Eloc = 0
        self.Enonloc = 0
        self.Ecoul = 0
        self.Exc = 0
        self.Eewald = 0
        self.Esic = 0

    @property
    def Etot(self):
        '''Total energy is the sum of all energy contributions.'''
        return self.Ekin + self.Eloc + self.Enonloc + self.Ecoul + self.Exc + self.Eewald + \
               self.Esic

    def __repr__(self):
        out = ''
        for ie in self.__slots__:
            energy = eval('self.' + ie)
            if energy != 0:
                out = f'{out}{ie.ljust(8)}: {energy:+.9f} Eh\n'
        out = f'{out}{"-" * 25}\nEtot    : {self.Etot:+.9f} Eh'
        return out


def get_Ekin(atoms, Y):
    '''Calculate the kinetic energy.

    Args:
        atoms :
            Atoms object.

        Y : array
            Expansion coefficients of orthogornal wave functions.

    Returns:
        Kinetic energy in Hartree as a float.
    '''
    # Arias: Ekin = -0.5 Tr(F Cdag L(C))
    return np.real(-0.5 * np.trace(np.diag(atoms.f) @ (Y.conj().T @ atoms.L(Y))))


def get_Ecoul(atoms, n):
    '''Calculate the coulomb energy.

    Args:
        atoms :
            Atoms object.

        n : array
            Real-space electronic density.

    Returns:
        Coulomb energy in Hartree as a float.
    '''
    # Arias: Ecoul = -(Jn)dag O(phi)
    phi = -4 * np.pi * atoms.Linv(atoms.O(atoms.J(n)))
    if atoms.cutcoul is None:
        return np.real(0.5 * n.conj().T @ atoms.Jdag(atoms.O(phi)))
    else:
        Rc = atoms.cutcoul
        correction = np.cos(np.sqrt(atoms.G2) * Rc) * atoms.O(phi)
        return np.real(0.5 * n.conj().T @ atoms.Jdag(atoms.O(phi) - correction))


def get_Exc(atoms, n, spinpol=False):
    '''Calculate the exchange-correlation energy.

    Args:
        atoms :
            Atoms object.

        n : array
            Real-space electronic density.

    Kwargs:
        spinpol : bool
            Choose if a spin-polarized exchange-correlation functional will be used.

    Returns:
        Exchange-correlation energy in Hartree as a float.
    '''
    # Arias: Exc = (Jn)dag O(J(exc))
    if atoms.spinpol or spinpol:
        exc = get_exc(atoms.exc, n, spinpol=True)[0]
    else:
        exc = get_exc(atoms.exc, n, spinpol=False)[0]
    return np.real(n.conj().T @ atoms.Jdag(atoms.O(atoms.J(exc))))


def get_Eloc(atoms, n):
    '''Calculate the local energy.

    Args:
        atoms :
            Atoms object.

        n : array
            Real-space electronic density.

    Returns:
        Local energy in Hartree as a float.
    '''
    return np.real(atoms.Vloc.conj().T @ n)


def get_Esic(atoms, n):
    '''Calculate the Perdew-Zunger self-interaction energy.'''
    # E_PZ-SIC = \sum_i Ecoul[n_i] + Exc[n_i, 0]
    Esic = 0
    for i in range(len(n)):
        # Normalize single-particle densities to 1
        n_tmp = n[i] / atoms.f[i]
        coul = get_Ecoul(atoms, n_tmp)
        # The exchange part for a SIC correction has to be spin polarized
        xc = get_Exc(atoms, n_tmp, spinpol=True)
        # SIC energy is scaled by the occupation
        Esic += (coul + xc) * atoms.f[i]
    return Esic


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/calc_energies.jl
def get_Enonloc(atoms, Y):
    '''Calculate the non-local energy.

    Args:
        atoms :
            Atoms object.

        Y : array
            Expansion coefficients of orthogornal wave functions.

    Returns:
        Non-local energy in Hartree as a float.
    '''
    Enonloc = 0
    if atoms.NbetaNL > 0:  # Only calculate non-local potential if necessary
        Natoms = atoms.Natoms
        Nstates = atoms.Ns
        prj2beta = atoms.prj2beta
        betaNL = atoms.betaNL

        betaNL_psi = np.dot(Y.T.conj(), betaNL).conj()

        for ist in range(Nstates):
            enl = 0
            for ia in range(Natoms):
                psp = atoms.GTH[atoms.atom[ia]]
                for l in range(psp['lmax']):
                    for m in range(-l, l + 1):
                        for iprj in range(psp['Nproj_l'][l]):
                            ibeta = prj2beta[iprj, ia, l, m + psp['lmax'] - 1] - 1
                            for jprj in range(psp['Nproj_l'][l]):
                                jbeta = prj2beta[jprj, ia, l, m + psp['lmax'] - 1] - 1
                                hij = psp['h'][l, iprj, jprj]
                                enl += hij * np.real(betaNL_psi[ist, ibeta].conj() *
                                       betaNL_psi[ist, jbeta])
            Enonloc += atoms.f[ist] * enl
    # We have to multiply with the cell volume, because of different orthogonalization
    return Enonloc * atoms.CellVol


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/calc_E_NN.jl
def get_Eewald(atoms, gcut=2, ebsl=1e-8):
    '''Calculate the Ewald energy.

    Args:
        atoms :
            Atoms object.

    Kwargs:
        gcut : float
            G-vector cut-off.

        ebsl : float
            Error tolerance

    Returns:
        Ewald energy in Hartree as a float.
    '''
    # For a plane wave code we have multiple contributions for the Ewald energy
    # Namely, a sum from contributions from real space, reciprocal space,
    # the self energy, (the dipole term [neglected]), and an additional electroneutrality term
    # See Eq. (4) https://juser.fz-juelich.de/record/16155/files/IAS_Series_06.pdf
    if atoms.cutcoul is not None:
        return 0

    Natoms = atoms.Natoms
    tau = atoms.X
    Zvals = atoms.Z
    omega = atoms.CellVol

    LatVecs = atoms.R
    t1 = LatVecs[0]
    t2 = LatVecs[1]
    t3 = LatVecs[2]
    t1m = np.sqrt(np.dot(t1, t1))
    t2m = np.sqrt(np.dot(t2, t2))
    t3m = np.sqrt(np.dot(t3, t3))

    RecVecs = 2 * np.pi * inv(LatVecs.conj().T)
    g1 = RecVecs[0]
    g2 = RecVecs[1]
    g3 = RecVecs[2]
    g1m = np.sqrt(np.dot(g1, g1))
    g2m = np.sqrt(np.dot(g2, g2))
    g3m = np.sqrt(np.dot(g3, g3))

    x = np.sum(Zvals**2)
    totalcharge = np.sum(Zvals)

    gexp = -np.log(ebsl)
    nu = 0.5 * np.sqrt(gcut**2 / gexp)

    tmax = np.sqrt(0.5 * gexp) / nu
    mmm1 = int(np.rint(tmax / t1m + 1.5))
    mmm2 = int(np.rint(tmax / t2m + 1.5))
    mmm3 = int(np.rint(tmax / t3m + 1.5))

    # Start by calculaton the self-energy
    Eewald = -nu * x / np.sqrt(np.pi)
    # Add the electroneutrality-term (Eq. 11)
    Eewald += -np.pi * totalcharge**2 / (2 * omega * nu**2)

    dtau = np.empty(3)
    G = np.empty(3)
    T = np.empty(3)
    for ia in range(Natoms):
        for ja in range(Natoms):
            dtau = tau[ia] - tau[ja]
            ZiZj = Zvals[ia] * Zvals[ja]
            for i in range(-mmm1, mmm1 + 1):
                for j in range(-mmm2, mmm2 + 1):
                    for k in range(-mmm3, mmm3 + 1):
                        if (ia != ja) or ((abs(i) + abs(j) + abs(k)) != 0):
                            T = i * t1 + j * t2 + k * t3
                            rmag = np.sqrt(np.sum((dtau - T)**2))
                            # Add the real space contribution
                            Eewald += 0.5 * ZiZj * erfc(rmag * nu) / rmag

    mmm1 = int(np.rint(gcut / g1m + 1.5))
    mmm2 = int(np.rint(gcut / g2m + 1.5))
    mmm3 = int(np.rint(gcut / g3m + 1.5))

    for ia in range(Natoms):
        for ja in range(Natoms):
            dtau = tau[ia] - tau[ja]
            ZiZj = Zvals[ia] * Zvals[ja]
            for i in range(-mmm1, mmm1 + 1):
                for j in range(-mmm2, mmm2 + 1):
                    for k in range(-mmm3, mmm3 + 1):
                        if (abs(i) + abs(j) + abs(k)) != 0:
                            G = i * g1 + j * g2 + k * g3
                            Gtau = np.sum(G * dtau)
                            G2 = np.sum(G**2)
                            # Add the reciprocal space contribution
                            x = 2 * np.pi / omega * np.exp(-0.25 * G2 / nu**2) / G2
                            Eewald += x * ZiZj * np.cos(Gtau)

    return Eewald



#!/usr/bin/env python3
'''
Various tools to check physical properties.
'''
import numpy as np

from .scf import get_epsilon


# Adapted from GPAW: https://gitlab.com/gpaw/gpaw/-/blob/master/gpaw/utilities/tools.py
def cutoff2gridspacing(E):
    '''Convert planewave energy cut-off to a real-space gridspacing using a.u.

    Args:
        E : float
            Energy in Hartree.

    Returns:
        Gridspacing in Bohr as a float.
    '''
    return np.pi / np.sqrt(2 * E)


# Adapted from GPAW: https://gitlab.com/gpaw/gpaw/-/blob/master/gpaw/utilities/tools.py
def gridspacing2cutoff(h):
    '''Convert real-space gridspacing to planewave energy cuto-ff using a.u.

    Args:
        h : float
            Gridspacing in Bohr.

    Returns:
        Cut-off in Hartree as a float.
    '''
    # In Hartree units, E=k^2/2, where k_max is approx. given by pi/h
    # See PRB, Vol 54, 14362 (1996)
    return 0.5 * (np.pi / h)**2


def center_of_mass(coords, masses=None):
    '''Calculate the center of mass for a set of coordinates and masses.

    Args:
        coords : array
            Array of real-space coordinates.

    Kwargs:
        masses : array
            Mass or weight for each coordinate.

    Returns:
        Center of mass as a float.
    '''
    if masses is None:
        masses = np.ones(len(coords))

    return np.sum(masses * coords.T, axis=1) / np.sum(masses)


def inertia_tensor(coords, masses=None):
    '''Calculate the inertia tensor for a set of coordinates and masses.

    Args:
        coords : array
            Array of real-space coordinates.

    Kwargs:
        masses : array
            Mass or weight for each coordinate.

    Returns:
        Inertia tensor as an array.
    '''
    if masses is None:
        masses = np.ones(len(coords))

    # For the definition see https://en.wikipedia.org/wiki/Moment_of_inertia#Definition_2
    I = np.empty((3, 3))
    I[0][0] = np.sum(masses * (coords[:, 1]**2 + coords[:, 2]**2))
    I[1][1] = np.sum(masses * (coords[:, 0]**2 + coords[:, 2]**2))
    I[2][2] = np.sum(masses * (coords[:, 0]**2 + coords[:, 1]**2))

    I[0][1] = I[1][0] = -np.sum(masses * (coords[:, 0] * coords[:, 1]))
    I[0][2] = I[2][0] = -np.sum(masses * (coords[:, 0] * coords[:, 2]))
    I[1][2] = I[2][1] = -np.sum(masses * (coords[:, 1] * coords[:, 2]))
    return I


def get_dipole(atoms):
    '''Calculate the electric dipole moment.

    Args:
        atoms :
            Atoms object.

    Returns:
        Electric dipole moment in e times Bohr as an array.
    '''
    # The dipole may be extremly large. This can be because of periodic boundary conditions.
    # E.g., the density gets "smeared" to the edges if the atom sits at one edge.
    # One fix can be to center the atom/molecule inside the box.
    n = atoms.n
    if n is None:
        print('ERROR: There is no density to calculate a dipole.')
        return 0

    # Diple moment: mu = \sum Z*X - \int n(r)*r dr
    mu = np.array([0, 0, 0], dtype=float)
    for i in range(atoms.Natoms):
        mu += atoms.Z[i] * atoms.X[i]

    dV = atoms.CellVol / np.prod(atoms.S)
    for dim in range(3):
        mu[dim] -= dV * np.sum(n * atoms.r[:, dim])
    return mu


def get_IP(atoms):
    '''Calculate the ionization potential by calculating the negative HOMO energy.

    Args:
        atoms :
            Atoms object.

    Returns:
        Ionization potential in Hartree as a float.
    '''
    epsilon = get_epsilon(atoms, atoms.W)
    IP = -epsilon[-1]
    return IP


def check_ortho(atoms, func):
    '''Check the orthogonality condition for a set of functions.

    Args:
        atoms :
            Atoms object.

        func : array
            Discretized set of functions.

    Returns:
        Orthogonality status of the set of functions as a bool.
    '''
    # Orthogonality condition: \int func1^* func2 dr = 0
    # Tolerance for the condition
    eps = 1e-9
    # It makes no sense to calculate anything for only one function
    if len(func) == 1:
        print('WARNING: Need at least two functions to check their orthogonality.')
        return True

    # We integrate over our unit cell, the integration borders then become a=0 and b=cell length
    # The integration prefactor dV is (b-a)/n, with n as the sampling
    # For a 3d integral we have to multiply for every direction
    dV = atoms.CellVol / np.prod(atoms.S)

    ortho_bool = True

    # Check the condition for every combination
    for i in range(func.shape[1]):
        for j in range(i + 1, func.shape[1]):
            res = dV * np.sum(func[:, i].conj() * func[:, j])
            tmp_bool = np.abs(res) < eps
            ortho_bool *= tmp_bool
            if atoms.verbose >= 3:
                print(f'Function {i} and {j}:\n\tValue: {res:.7f}\n\tOrthogonal: {tmp_bool}')
    print(f'Orthogonal: {ortho_bool}')
    return ortho_bool


def check_norm(atoms, func):
    '''Check the normalization condition for a set of functions.

    Args:
        atoms :
            Atoms object.

        func : array
            Discretized set of functions.

    Returns:
        Normalization status of the set of functions as a bool.
    '''
    # Normality condition: \int func^* func dr = 1
    # Tolerance for the condition
    eps = 1e-9
    # We integrate over our unit cell, the integration borders then become a=0 and b=cell length
    # The integration prefactor dV is (b-a)/n, with n as the sampling
    # For a 3d integral we have to multiply for every direction
    dV = atoms.CellVol / np.prod(atoms.S)

    norm_bool = True

    # Check the condition for every function
    for i in range(func.shape[1]):
        res = dV * np.sum(func[:, i].conj() * func[:, i])
        tmp_bool = np.abs(1 - res) < eps
        norm_bool *= tmp_bool
        if atoms.verbose >= 3:
            print(f'Function {i}:\n\tValue: {res:.7f}\n\tNormalized: {tmp_bool}')
    print(f'Normalized: {norm_bool}')
    return norm_bool


def check_orthonorm(atoms, func):
    '''Check the orthonormality conditions for a set of functions.

    Args:
        atoms :
            Atoms object.

        func : array
            Discretized set of functions.

    Returns:
        Orthonormality status of the set of functions as a bool.
    '''
    ortho_bool = check_ortho(atoms, func)
    norm_bool = check_norm(atoms, func)
    print(f'Orthonormal: {ortho_bool * norm_bool}')
    return ortho_bool * norm_bool



#!/usr/bin/env python3
'''
Workflow functions that combine functions to get one property.
'''
from eminus.atoms_io import write_cube
from eminus.localizer import get_FLOs, get_FOs
from eminus.scf import get_psi
from .fods import get_fods, remove_core_fods


def KSO(atoms, write_cubes=True, **kwargs):
    '''Generate Kohn-Sham orbitals and optionally save them as cube files.

    Args:
        atoms :
            Atoms object.

    Kwargs:
        write_cubes : bool
            Write orbitals to cube files.

    Returns:
        Kohn-Sham orbitals as an array.
    '''
    KSOs = get_psi(atoms, atoms.W)
    name = ''
    for ia in set(atoms.atom):
        name += f'{ia}{atoms.atom.count(ia)}'
    if write_cubes:
        for i in range(atoms.Ns):
            write_cube(atoms, KSOs[:, i], f'{name}_KSO_{i}.cube')
    return KSOs


def FO(atoms, write_cubes=True, fods=None):
    '''Generate Fermi orbitals and optionally save them as cube files.

    Args:
        atoms :
            Atoms object.

    Kwargs:
        write_cubes : bool
            Write orbitals to cube files.

        fods : array
            Fermi-orbital descriptors.

    Returns:
        Fermi orbitals as an array.
    '''
    KSOs = get_psi(atoms, atoms.W)
    if fods is None:
        fods = get_fods(atoms)
    fods = remove_core_fods(atoms, fods)
    FOs = get_FOs(atoms, KSOs, fods)
    name = ''
    for ia in set(atoms.atom):
        name += f'{ia}{atoms.atom.count(ia)}'
    if write_cubes:
        for i in range(atoms.Ns):
            write_cube(atoms, FOs[:, i], f'{name}_FO_{i}.cube')
    return FOs


def FLO(atoms, write_cubes=True, fods=None):
    '''Generate Fermi-Löwdin orbitals and optionally save them as cube files.

    Args:
        atoms :
            Atoms object.

    Kwargs:
        write_cubes : bool
            Write orbitals to cube files.

        fods : array
            Fermi-orbital descriptors.

    Returns:
        Fermi-Löwdin orbitals as an array.
    '''
    KSOs = get_psi(atoms, atoms.W)
    if fods is None:
        fods = get_fods(atoms)
    fods = remove_core_fods(atoms, fods)
    FLOs = get_FLOs(atoms, KSOs, fods)
    name = ''
    for ia in set(atoms.atom):
        name += f'{ia}{atoms.atom.count(ia)}'
    if write_cubes:
        for i in range(atoms.Ns):
            write_cube(atoms, FLOs[:, i], f'{name}_FLO_{i}.cube')
    return FLOs



#!/usr/bin/env python3
'''
SCF calculation functionalities with relevant sub-function.
'''
from timeit import default_timer

import numpy as np
from numpy.linalg import eig, inv, norm
from numpy.random import randn, seed
from scipy.linalg import sqrtm

from .energies import get_Ecoul, get_Eewald, get_Ekin, get_Eloc, get_Enonloc, get_Exc
from .exc import get_exc
from .gth import calc_Vnonloc
from .utils import Diagprod, dotprod


def SCF(atoms, guess='gaussian', etol=1e-7, min=None, cgform=1):
    '''SCF function to handle direct minimizations.

    Args:
        atoms :
            Atoms object.

    Kwargs:
        guess : str
            Initial guess method for the basis functions (case insensitive).
            Example: 'Gauss'; 'gaussian'; 'random'; 'rand'; 'Experimental'; 'exp'
            Default: 'gaussian'

        etol : float
            Convergence tolerance of the total energy.
            Default: 1e-7

        min : dict
            Dictionary to set the maximum amount of steps per minimization method and their order.
            Example: {'sd': 10, 'pccg': 100}; {'pccg': 10, 'lm': 25, 'pclm': 50}
            Default: None (will default to {'pccg': 100})

        cgform : int
            Conjugated-gradient from for the preconditioned conjugate-gradient minimization (pccg).
            1 for Fletcher-Reeves, 2 for Polak-Ribiere, and 3 for Hestenes-Stiefel.
            Default: 1

    Returns:
        Total energy as a float.
    '''
    # Map minimization names and functions, also use this dict to save times and iterations
    minimizer = {
        'sd': {
            'func': sd,
            'name': 'steepest descent minimization'
        },
        'lm': {
            'func': lm,
            'name': 'line minimization'
        },
        'pclm': {
            'func': pclm,
            'name': 'preconditioned line minimization'
        },
        'pccg': {
            'func': pccg,
            'name': 'preconditioned conjugate-gradient minimization'
        }
    }

    # Set min here, better not use mutable data types in signatures
    if min is None:
        min = {'pccg': 100}

    # Print some useful informations
    if atoms.verbose >= 3:
        print(f'--- System informations ---\n{atoms}\n')
    if atoms.verbose >= 4:
        print(f'--- Cell informations ---\nSide lengths: {atoms.a} Bohr\n'
              f'Cut-off energy: {atoms.ecut} Hartree\n'
              f'Sampling per axis: ({atoms.S[0]}, {atoms.S[1]}, {atoms.S[2]})\n')
        print(f'--- Calculation informations ---\nSpin polarization: {atoms.spinpol}\n'
              f'Number of states: {atoms.Ns}\n'
              f'Occupation per state: {atoms.f}\n'
              f'Potential: {atoms.pot}\n'
              f'Non-local contribution: {atoms.NbetaNL > 0}\n'
              f'Coulomb-truncation: {atoms.cutcoul is not None}\n'
              f'XC functionals: {atoms.exc}\n'
              f'Compression: {len(atoms.G2) / len(atoms.G2c):.5f}\n')

    # Set up basis functions
    guess = guess.lower()
    if guess in ('gauss', 'gaussian'):
        # Start with gaussians at atom positions
        W = guess_gaussian(atoms)
    elif guess in ('rand', 'random'):
        # Start with randomized, complex basis functions with a random seed
        W = guess_random(atoms, complex=True, reproduce=True)
    elif guess in ('exp', 'experimental'):
        # Start with s-type orbitals. This may work good for small systems.
        W = guess_experimental(atoms)

    # Calculate ewald energy
    atoms.energies.Eewald = get_Eewald(atoms)

    # Start minimization procedures
    if atoms.verbose >= 3:
        print('--- SCF data ---')
    Etots = []
    for imin in min:
        start = default_timer()
        if atoms.verbose >= 2:
            print(f'Start {minimizer[imin]["name"]}...')
        W, Elist = minimizer[imin]['func'](atoms, W, min[imin], etol, cgform=cgform)
        end = default_timer()
        minimizer[imin]['time'] = end - start
        minimizer[imin]['iteration'] = len(Elist)
        Etots += Elist
    if atoms.verbose >= 1:
        if abs(Etots[-2] - Etots[-1]) < etol:
            print(f'SCF converged after {len(Etots)} iterations.')
        else:
            print('SCF not converged!')

    # Print SCF data
    if atoms.verbose >= 2:
        if atoms.verbose >= 3:
            print(f'\n--- SCF results ---\nStarting guess: {guess}')
        t_total = 0
        for imin in min:
            N = minimizer[imin]['iteration']
            t = minimizer[imin]['time']
            t_total += t
            if atoms.verbose >= 4:
                print(f'\nMinimizer: {imin}'
                      f'\nIterations:\t{N}'
                      f'\nTime:\t\t{t:.5f}s'
                      f'\nTime/Iteration:\t{t / N:.5f}s')
        print(f'Total SCF time: {t_total:.5f}s')

    # Print energy data
    if atoms.verbose >= 3:
        print('\n--- Energy data ---')
    if atoms.verbose >= 4:
        print(f'{atoms.energies}')
    elif atoms.verbose >= 1:
        print(f'Total energy: {atoms.energies.Etot:.9f} Eh')

    # Save basis functions and density to reuse them later
    atoms.W = orth(atoms, W)
    atoms.n = get_n_total(atoms, atoms.W)
    return atoms.energies.Etot


def H(atoms, W, n=None):
    '''Left-hand side of the eigenvalue equation.'''
    Y = orth(atoms, W)  # Orthogonalize at the start
    if n is None:
        n = get_n_total(atoms, Y)
    phi = -4 * np.pi * atoms.Linv(atoms.O(atoms.J(n)))

    # Calculate the exchange-correlation potential
    # We get the full potential in the functinal definition, unlike in the Arias notation
    vxc = get_exc(atoms.exc, n, atoms.spinpol)[1]
    Vxc = atoms.Jdag(atoms.O(atoms.J(vxc)))

    # Calculate the effective potential, with or without Coulomb truncation
    if atoms.cutcoul is None:
        Veff = atoms.Vloc + Vxc + atoms.Jdag(atoms.O(phi))
    else:
        Rc = atoms.cutcoul
        correction = np.cos(np.sqrt(atoms.G2) * Rc) * atoms.O(phi)
        Veff = atoms.Vloc + Vxc + atoms.Jdag(atoms.O(phi) - correction)

    Vkin_psi = -0.5 * atoms.L(W)
    Vnonloc_psi = calc_Vnonloc(atoms, W)
    return Vkin_psi + atoms.Idag(Diagprod(Veff, atoms.I(W))) + Vnonloc_psi


def Q(inp, U):
    '''Operator needed to calculate gradients with non-constant occupations.'''
    mu, V = eig(U)
    mu = np.reshape(mu, (len(mu), 1))
    denom = np.sqrt(mu) @ np.ones((1, len(mu)))
    denom = denom + denom.conj().T
    return V @ ((V.conj().T @ inp @ V) / denom) @ V.conj().T


def get_E(atoms, W):
    '''Calculate all the energy contributions.'''
    Y = orth(atoms, W)
    n = get_n_total(atoms, Y)
    atoms.energies.Ekin = get_Ekin(atoms, Y)
    atoms.energies.Eloc = get_Eloc(atoms, n)
    atoms.energies.Enonloc = get_Enonloc(atoms, Y)
    atoms.energies.Ecoul = get_Ecoul(atoms, n)
    atoms.energies.Exc = get_Exc(atoms, n)
    return atoms.energies.Etot


def get_grad(atoms, W):
    '''Calculate the energy gradient with respect to W.'''
    U = W.conj().T @ atoms.O(W)
    invU = inv(U)
    HW = H(atoms, W)
    F = np.diag(atoms.f)
    U12 = sqrtm(inv(U))
    Ht = U12 @ (W.conj().T @ HW) @ U12
    return (HW - (atoms.O(W) @ invU) @ (W.conj().T @ HW)) @ (U12 @ F @ U12) + \
           atoms.O(W) @ (U12 @ Q(Ht @ F - F @ Ht, U))


def check_energies(atoms, Elist, etol, linmin=None, cg=None):
    '''Check the energies for every SCF cycle and handle the output.'''
    Nit = len(Elist)

    # Output handling
    if linmin is not None:
        linmin = f'\tlinmin-test: {linmin:+.7f}'
    else:
        linmin = ''
    if cg is not None:
        cg = f'\tcg-test: {cg:+.7f}'
    else:
        cg = ''

    if atoms.verbose >= 5:
        print(f'Iteration: {Nit} {linmin} {cg}\n{atoms.energies}\n')
    elif atoms.verbose >= 4:
        print(f'Iteration: {Nit}  \tEtot: {atoms.energies.Etot:+.7f} {linmin} {cg}')
    elif atoms.verbose >= 3:
        print(f'Iteration: {Nit}  \tEtot: {atoms.energies.Etot:+.7f}')

    # Check for convergence
    if Nit > 1:
        if abs(Elist[-2] - Elist[-1]) < etol:
            return True
        if Elist[-1] > Elist[-2] and atoms.verbose >= 3:
            print('WARNING: Total energy is not decreasing.')
    return False


def sd(atoms, W, Nit, etol, **kwargs):
    '''Steepest descent minimization algorithm.'''
    Elist = []
    alpha = 3e-5

    for _ in range(Nit):
        W = W - alpha * get_grad(atoms, W)
        E = get_E(atoms, W)
        Elist.append(E)
        if check_energies(atoms, Elist, etol):
            break
    return W, Elist


def lm(atoms, W, Nit, etol, **kwargs):
    '''Line minimization algorithm.'''
    Elist = []
    alphat = 3e-5

    # Do the first step without the linmin test
    g = get_grad(atoms, W)
    d = -g
    gt = get_grad(atoms, W + alphat * d)
    alpha = alphat * dotprod(g, d) / dotprod(g - gt, d)
    W = W + alpha * d
    E = get_E(atoms, W)
    Elist.append(E)
    check_energies(atoms, Elist, etol)

    for _ in range(1, Nit):
        g = get_grad(atoms, W)
        linmin = dotprod(g, d) / np.sqrt(dotprod(g, g) * dotprod(d, d))
        d = -g
        gt = get_grad(atoms, W + alphat * d)
        # The denominator can get zero, add this check to prevent this
        denom = dotprod(g - gt, d)
        if abs(denom) > 1e-16:  # 1e-16 is the range under float64 machine precision
            alpha = alphat * dotprod(g, d) / denom
        else:
            alpha = alphat * dotprod(g, d) / 1e-16
        W = W + alpha * d
        E = get_E(atoms, W)
        Elist.append(E)
        if check_energies(atoms, Elist, etol, linmin):
            break
    return W, Elist


def pclm(atoms, W, Nit, etol, **kwargs):
    '''Preconditioned line minimization algorithm.'''
    Elist = []
    alphat = 3e-5

    # Do the first step without the linmin test
    g = get_grad(atoms, W)
    d = -atoms.K(g)
    gt = get_grad(atoms, W + alphat * d)
    alpha = alphat * dotprod(g, d) / dotprod(g - gt, d)
    W = W + alpha * d
    E = get_E(atoms, W)
    Elist.append(E)
    check_energies(atoms, Elist, etol)

    for _ in range(1, Nit):
        g = get_grad(atoms, W)
        linmin = dotprod(g, d) / np.sqrt(dotprod(g, g) * dotprod(d, d))
        d = -atoms.K(g)
        gt = get_grad(atoms, W + alphat * d)
        # The denominator can get zero, add this check to prevent this
        denom = dotprod(g - gt, d)
        if abs(denom) > 1e-16:  # 1e-16 is the range under float64 machine precision
            alpha = alphat * dotprod(g, d) / denom
        else:
            alpha = alphat * dotprod(g, d) / 1e-16
        W = W + alpha * d
        E = get_E(atoms, W)
        Elist.append(E)
        if check_energies(atoms, Elist, etol, linmin):
            break
    return W, Elist


def pccg(atoms, W, Nit, etol, cgform=1):
    '''Preconditioned conjugate-gradient algorithm.'''
    Elist = []
    alphat = 3e-5

    # Do the first step without the linmin and cg test
    g = get_grad(atoms, W)
    d = -atoms.K(g)
    gt = get_grad(atoms, W + alphat * d)
    alpha = alphat * dotprod(g, d) / dotprod(g - gt, d)
    W = W + alpha * d
    dold = d
    gold = g
    E = get_E(atoms, W)
    Elist.append(E)
    check_energies(atoms, Elist, etol)

    for _ in range(1, Nit):
        g = get_grad(atoms, W)
        linmin = dotprod(g, dold) / np.sqrt(dotprod(g, g) * dotprod(dold, dold))
        cg = dotprod(g, atoms.K(gold)) / np.sqrt(dotprod(g, atoms.K(g)) *
             dotprod(gold, atoms.K(gold)))
        if cgform == 1:  # Fletcher-Reeves
            beta = dotprod(g, atoms.K(g)) / dotprod(gold, atoms.K(gold))
        elif cgform == 2:  # Polak-Ribiere
            beta = dotprod(g - gold, atoms.K(g)) / dotprod(gold, atoms.K(gold))
        elif cgform == 3:  # Hestenes-Stiefel
            beta = dotprod(g - gold, atoms.K(g)) / dotprod(g - gold, dold)
        d = -atoms.K(g) + beta * dold
        gt = get_grad(atoms, W + alphat * d)
        # The denominator can get zero, add this check to prevent this
        denom = dotprod(g - gt, d)
        if abs(denom) > 1e-16:  # 1e-16 is the range under float64 machine precision
            alpha = alphat * dotprod(g, d) / denom
        else:
            alpha = alphat * dotprod(g, d) / 1e-16
        W = W + alpha * d
        dold = d
        gold = g
        E = get_E(atoms, W)
        Elist.append(E)
        if check_energies(atoms, Elist, etol, linmin, cg):
            break
    return W, Elist


def orth(atoms, W):
    '''Orthogonalize coefficent matrix W.'''
    return W @ inv(sqrtm(W.conj().T @ atoms.O(W)))


def get_psi(atoms, Y, n=None):
    '''Calculate eigenstates from H.'''
    mu = Y.conj().T @ H(atoms, Y, n)
    _, D = eig(mu)
    return Y @ D


def get_epsilon(atoms, Y, n=None):
    '''Calculate eigenvalues from H.'''
    mu = Y.conj().T @ H(atoms, Y, n)
    epsilon, _ = eig(mu)
    epsilon = np.real(epsilon)
    return np.sort(epsilon)


def get_n_total(atoms, Y):
    '''Calculate the total electronic density.'''
    Y = Y.T
    n = np.zeros((np.prod(atoms.S), 1))
    for i in range(Y.shape[0]):
        psi = atoms.I(Y[i])
        n += atoms.f[i] * np.real(psi.conj() * psi)
    return n.T[0]


def get_n_single(atoms, Y):
    '''Calculate the single electronic densities.'''
    Y = Y.T
    n = np.empty((np.prod(atoms.S), len(Y)))
    for i in range(Y.shape[0]):
        psi = atoms.I(Y[i])
        n[:, i] = atoms.f[i] * np.real(psi.conj() * psi).T
    return n.T


def guess_random(atoms, complex=True, reproduce=False):
    '''Generate random coefficents as starting values.'''
    if reproduce:
        seed(42)
    if complex:
        W = randn(len(atoms.active[0]), atoms.Ns) + 1j * randn(len(atoms.active[0]), atoms.Ns)
    else:
        W = randn(len(atoms.active[0]), atoms.Ns)
    return orth(atoms, W)


def guess_gaussian(atoms):
    '''Generate inital-guess coefficents using normalized Gaussians as starting values.'''
    # Start with a randomized basis set
    W = guess_random(atoms, complex=True, reproduce=True)
    W = orth(atoms, W)

    eps = 1e-9
    sigma = 0.5
    normal = (2 * np.pi * sigma**2)**(3 / 2)

    # Calculate a density from normalized Gauss functions
    n = np.zeros(len(atoms.r)) + eps  # We divide by n in exc functionals, so 0 would be bad
    for ia in range(atoms.Natoms):
        r = norm(atoms.r - atoms.X[ia], axis=1)
        n += atoms.Z[ia] * np.exp(-r**2 / (2 * sigma**2)) / normal

    # Calculate the eigenfunctions
    W = get_psi(atoms, W, n)
    return W


def guess_experimental(atoms):
    '''Generate a s-type basis set.'''
    eps = 1e-9
    sigma = 0.5
    normal = (2 * np.pi * sigma**2)**(3 / 2)

    W = np.empty((len(atoms.r), atoms.Ns))
    for ist in range(atoms.Ns):
        # If we have more states than atoms, start all over again
        ia = ist % atoms.Natoms
        r = norm(atoms.r - atoms.X[ia], axis=1)
        W[:, ist] = atoms.Z[ia] * np.exp(-r**2 / (2 * sigma**2)) / normal + eps
        # Vary each W minimally by eps with changing signs, such that Wdag*O(W) is not singulary
        eps *= -1
    # Transform from real-space to reciprocal space
    # There is no transformation on the active space for this, so do it "manually"
    W = atoms.J(W)[atoms.active]
    return orth(atoms, W)



#!/usr/bin/env python3
'''
Collection of miscellaneous potentials.
'''
import numpy as np
from numpy.linalg import norm


def Harmonic(atoms):
    '''Harmonical potential. Can be used for quantum dot calculations.

    Args:
        atoms :
            Atoms object.

    Returns:
        Harmonical potential as an array.
    '''
    dr = norm(atoms.r - np.sum(atoms.R, axis=1) / 2, axis=1)
    V = 2 * dr**2
    return atoms.Jdag(atoms.O(atoms.J(V)))


def Coulomb(atoms):
    '''Coulomb potential. Can be used for Hydrogen calculations.

    Args:
        atoms :
            Atoms object.

    Returns:
        Coulomb potential as an array.
    '''
    Z = atoms.Z[0]  # Potential should only be used for same species
    Vps = -4 * np.pi * Z / atoms.G2[1:]
    Vps = np.concatenate(([0], Vps))

    Sf = np.sum(atoms.Sf, axis=0)
    return atoms.J(Vps * Sf)


def Ge(atoms):
    '''Starkloff-Joannopoulos pseudopotential for Germanium. Fourier transformed by Tomas Arias.

    Args:
        atoms :
            Atoms object.

    Returns:
        Germanium pseudopotential as an array.
    '''
    Z = atoms.Z[0]  # Potential should only be used for same species/Germanium
    lamda = 18.5
    rc = 1.052
    Gm = np.sqrt(atoms.G2[1:])

    Vps = -2 * np.pi * np.exp(-np.pi * Gm / lamda) * np.cos(rc * Gm) * (Gm / lamda) / \
          (1 - np.exp(-2 * np.pi * Gm / lamda))
    for n in range(5):
        Vps += (-1)**n * np.exp(-lamda * rc * n) / (1 + (n * lamda / Gm)**2)
    Vps = Vps * 4 * np.pi * Z / Gm**2 * (1 + np.exp(-lamda * rc)) - 4 * np.pi * Z / Gm**2

    n = np.arange(1, 5)
    eps = 4 * np.pi * Z * (1 + np.exp(-lamda * rc)) * (rc**2 / 2 + 1 / lamda**2 * (np.pi**2 / 6 +
          np.sum((-1)**n * np.exp(-lamda * rc * n) / n**2)))
    Vps = np.concatenate(([eps], Vps))

    Sf = np.sum(atoms.Sf, axis=0)
    return atoms.J(Vps * Sf)


def init_pot(atoms):
    '''Handle and initialize potentials.

    Args:
        atoms :
            Atoms object.

    Returns:
        Potential as an array.
    '''
    implemented = {'harmonic': Harmonic, 'coulomb': Coulomb, 'ge': Ge}
    try:
        pot = implemented[atoms.pot]
    except KeyError:
        print(f'ERROR: No potential found for "{atoms.pot}"')
    return pot(atoms)



#!/usr/bin/env python3
'''
eminus - A plane wave density funtional theory code.

Minimal usage example to do a DFT calculation for Hydrogen::

    from eminus import *
    atoms = Atoms('H', [0, 0, 0])
    SCF(atoms)
'''
from .atoms import Atoms
from .atoms_io import load_atoms, read_cube, read_xyz, save_atoms, write_cube, write_xyz
from .scf import get_epsilon, get_psi, SCF
from .version import info, __version__

__all__ = ['Atoms', 'get_epsilon', 'get_psi', 'info', 'load_atoms', 'read_cube', 'read_xyz',
           'save_atoms', 'SCF', 'write_cube', 'write_xyz', '__version__']



#!/usr/bin/env python3
'''
Viewer functions for Jupyter notebooks.
'''
import numpy as np
from numpy.linalg import norm
try:
    from nglview import NGLWidget, TextStructure
    from vispy import scene
except ImportError:
    print('ERROR: Necessary addon dependecies not found. '
          'To use this module, install the package with addons, e.g., with '
          '"pip install eminus[addons]"')

from eminus.atoms_io import create_pdb, read_cube, read_xyz
from .fods import split_atom_and_fod


# Adapted from https://github.com/MolSSI/QCFractal/issues/374
def view_mol(filename, isovalue=0.01, **kwargs):
    '''Display molecules and orbitals.

    Args:
        filename : str
            Input file path/name. This can be either a cube or xyz file.

    Kwargs:
        isovalue : float
            Isovalue for sizing orbitals.

    Returns:
        Viewable object as a NGLWidget.
    '''
    if isinstance(isovalue, str):
        isovalue = float(isovalue)
    view = NGLWidget(**kwargs)
    view._set_size('400px', '400px')

    if filename.endswith('.xyz'):
        # Atoms
        atom, X = read_xyz(filename)
        atom, X, X_fod = split_atom_and_fod(atom, X)
        view.add_component(TextStructure(create_pdb(atom, X)))
        view[0].clear()
        view[0].add_ball_and_stick()
        # FODs
        if len(X_fod) > 0:
            view.add_component(TextStructure(create_pdb(['X'] * len(X_fod), X_fod)))
            view[1].clear()
            view[1].add_ball_and_stick('_X', color='red', radius=0.1)
        view.center()

    if filename.endswith('.cube'):
        # Atoms and unit cell
        atom, X, _, a, _ = read_cube(filename)
        atom, X, X_fod = split_atom_and_fod(atom, X)
        view.add_component(TextStructure(create_pdb(atom, X, a)))
        view[0].clear()
        view[0].add_ball_and_stick()
        view.add_unitcell()
        view.center()
        # Spin up
        view.add_component(filename)
        view[1].clear()
        # Negate isovalue here as a workaround for some display bugs
        view[1].add_surface(negateIsolevel=True,
                            isolevelType='value',
                            isolevel=-isovalue,
                            color='lightgreen',
                            opacity=0.75,
                            side='front')
        # Spin down
        view.add_component(filename)
        view[2].clear()
        view[2].add_surface(negateIsolevel=True,
                            isolevelType='value',
                            isolevel=isovalue,
                            color='red',
                            opacity=0.75,
                            side='front')
        # FODs
        if len(X_fod) > 0:
            view.add_component(TextStructure(create_pdb(['X'] * len(X_fod), X_fod)))
            view[3].clear()
            view[3].add_ball_and_stick('_X', color='red', radius=0.1)
    return view


def view_grid(coords, extra=None):
    '''Display 3D-coordinates, e.g., grid points.

    Args:
        coords : array
            Grid points.

    Kwargs:
        extra : array
            Extra coordinates to display.

    Returns:
        Viewable object as a SceneCanvas.
    '''
    # Set up view
    canvas = scene.SceneCanvas(keys='interactive', show=True, size=(400, 400))
    view = canvas.central_widget.add_view()
    view.size = (400, 400)
    view.camera = 'arcball'
    view.camera.center = (np.mean(coords[:, 0]), np.mean(coords[:, 1]), np.mean(coords[:, 2]))
    view.camera.distance = np.max(norm(coords, axis=1)) * 2

    # Add data
    grid = scene.visuals.Markers()
    grid.set_data(coords, face_color='lightgreen', edge_width=0, size=2)
    view.add(grid)
    if extra is not None:
        grid_extra = scene.visuals.Markers()
        grid_extra.set_data(extra, face_color='red', edge_width=0, size=8)
        view.add(grid_extra)

    canvas.app.run()
    return canvas



#!/usr/bin/env python3
'''
Utilities to use Goedecker, Teter, and Hutter pseudopotentials.
'''
from glob import glob
from os.path import basename

import numpy as np

from . import __path__
from .utils import Ylm_real


def init_gth_loc(atoms):
    '''Initialize parameters to calculate local contributions of GTH pseudopotentials.

    Args:
        atoms :
            Atoms object.

    Returns:
        Local GTH potential contribution as an array.
    '''
    G2 = atoms.G2
    atom = atoms.atom
    species = set(atom)

    Vsp = np.empty(len(G2))  # Potential for every species
    Vloc = np.zeros(len(G2))  # Total local potential
    for isp in species:
        psp = atoms.GTH[isp]
        rloc = psp['rlocal']
        Zion = psp['Zval']
        C1 = psp['C'][0]
        C2 = psp['C'][1]
        C3 = psp['C'][2]
        C4 = psp['C'][3]

        omega = 1  # Normally this would be det(atoms.R), but Arias notation is off by this factor
        rlocG2 = G2[1:] * rloc**2

        Vsp[1:] = -4 * np.pi * Zion / omega * np.exp(-0.5 * rlocG2) / G2[1:] + \
                  np.sqrt((2 * np.pi)**3) * rloc**3 / omega * np.exp(-0.5 * rlocG2) * \
                  (C1 + C2 * (3 - rlocG2) + C3 * (15 - 10 * rlocG2 + rlocG2**2) +
                  C4 * (105 - 105 * rlocG2 + 21 * rlocG2**2 - rlocG2**3))
        Vsp[0] = 2 * np.pi * Zion * rloc**2 + \
                 (2 * np.pi)**1.5 * rloc**3 * (C1 + 3 * C2 + 15 * C3 + 105 * C4)

        # Sum up the structure factor for every species
        Sf = np.zeros(len(atoms.Sf[0]), dtype=complex)
        for ia in range(len(atom)):
            if atom[ia] == isp:
                Sf += atoms.Sf[ia]
        Vloc += np.real(atoms.J(Vsp * Sf))
    return Vloc


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/PsPotNL.jl
def init_gth_nonloc(atoms):
    '''Initialize parameters to calculate non-local contributions of GTH pseudopotentials.

    Args:
        atoms :
            Atoms object.

    Returns:
        NbetaNL, prj2beta, and betaNL as a tuple(int, array, array).
    '''
    Natoms = atoms.Natoms
    Npoints = len(atoms.active[0])
    CellVol = atoms.CellVol

    prj2beta = np.zeros([3, Natoms, 4, 7], dtype=int)
    prj2beta[:] = -1  # Set to invalid index

    NbetaNL = 0
    for ia in range(Natoms):
        psp = atoms.GTH[atoms.atom[ia]]
        for l in range(psp['lmax']):
            for m in range(-l, l + 1):
                for iprj in range(psp['Nproj_l'][l]):
                    NbetaNL += 1
                    prj2beta[iprj, ia, l, m + psp['lmax'] - 1] = NbetaNL

    g = atoms.Gc  # Simplified, would normally be G+k
    Gm = np.sqrt(atoms.G2c)

    ibeta = 0
    betaNL = np.zeros([Npoints, NbetaNL], dtype=complex)
    for ia in range(Natoms):
        Sf = atoms.Idag(atoms.J(atoms.Sf[ia]))
        psp = atoms.GTH[atoms.atom[ia]]
        for l in range(psp['lmax']):
            for m in range(-l, l + 1):
                for iprj in range(psp['Nproj_l'][l]):
                    betaNL[:, ibeta] = (-1j)**l * Ylm_real(l, m, g) * \
                                       eval_proj_G(psp, l, iprj + 1, Gm, CellVol) * Sf
                    ibeta += 1
    return NbetaNL, prj2beta, betaNL


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/op_V_Ps_nloc.jl
def calc_Vnonloc(atoms, W):
    '''Calculate the non-local pseudopotential, applied on the basis functions W.

    Args:
        atoms :
            Atoms object.

        W : array
            Expansion coefficients of unconstrained wave functions.

    Returns:
        Non-local GTH potential contribution as an array.
    '''
    Npoints = len(W)
    Nstates = atoms.Ns

    Vpsi = np.zeros([Npoints, Nstates], dtype=complex)
    if atoms.NbetaNL > 0:  # Only calculate non-local potential if necessary
        Natoms = atoms.Natoms
        prj2beta = atoms.prj2beta
        betaNL = atoms.betaNL

        betaNL_psi = np.dot(W.T.conj(), betaNL).conj()

        for ist in range(Nstates):
            for ia in range(Natoms):
                psp = atoms.GTH[atoms.atom[ia]]
                for l in range(psp['lmax']):
                    for m in range(-l, l + 1):
                        for iprj in range(psp['Nproj_l'][l]):
                            ibeta = prj2beta[iprj, ia, l, m + psp['lmax'] - 1] - 1
                            for jprj in range(psp['Nproj_l'][l]):
                                jbeta = prj2beta[jprj, ia, l, m + psp['lmax'] - 1] - 1
                                hij = psp['h'][l, iprj, jprj]
                                Vpsi[:, ist] += hij * betaNL[:, ibeta] * betaNL_psi[ist, jbeta]
    # We have to multiply with the cell volume, because of different orthogonalization
    return Vpsi * atoms.CellVol


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/PsPot_GTH.jl
def eval_proj_G(psp, l, iprj, Gm, CellVol):
    '''Evaluate GTH projector functions in G-space.

    Args:
        psp : dict
            GTH parameters.

        l : int
            Angular momentum number.

        iprj : int
            Nproj_l index.

        Gm : array
            Magnitude of G-vectors.

        CellVol : float
            Unit cell volume.

    Returns:
        GTH projector as an array.
    '''
    rrl = psp['rc'][l]
    Gr2 = (Gm * rrl)**2

    if l == 0:  # s-channel
        if iprj == 1:
            Vprj = np.exp(-0.5 * Gr2)
        elif iprj == 2:
            Vprj = 2 / np.sqrt(15) * np.exp(-0.5 * Gr2) * (3 - Gr2)
        elif iprj == 3:
            Vprj = (4 / 3) / np.sqrt(105) * np.exp(-0.5 * Gr2) * (15 - 10 * Gr2 + Gr2**2)
    elif l == 1:  # p-channel
        if iprj == 1:
            Vprj = (1 / np.sqrt(3)) * np.exp(-0.5 * Gr2) * Gm
        elif iprj == 2:
            Vprj = (2 / np.sqrt(105)) * np.exp(-0.5 * Gr2) * Gm * (5 - Gr2)
        elif iprj == 3:
            Vprj = (4 / 3) / np.sqrt(1155) * np.exp(-0.5 * Gr2) * Gm * (35 - 14 * Gr2 + Gr2**2)
    elif l == 2:  # d-channel
        if iprj == 1:
            Vprj = (1 / np.sqrt(15)) * np.exp(-0.5 * Gr2) * Gm**2
        elif iprj == 2:
            Vprj = (2 / 3) / np.sqrt(105) * np.exp(-0.5 * Gr2) * Gm**2 * (7 - Gr2)
    elif l == 3:  # f-channel
        # Only one projector
        Vprj = Gm**3 * np.exp(-0.5 * Gr2) / np.sqrt(105)
    else:
        print(f'ERROR: No projector found for l={l}')

    pre = 4 * np.pi**(5 / 4) * np.sqrt(2**(l + 1) * rrl**(2 * l + 3) / CellVol)
    return pre * Vprj


def read_gth(system, charge=None, psp_path=None):
    '''Read GTH files for a given system.

    Args:
        system : str
            Atom name.

    Kwargs:
        charge : int
            Valence charge.

        psp_path : str
            Path to GTH pseudopotential files. None will default to /installation_path/pade_gth/.

    Returns:
        GTH parameters as a dictionary.
    '''
    if psp_path is None:
        psp_path = f'{__path__[0]}/pade_gth/'

    if charge is not None:
        f_psp = f'{psp_path}{system}-q{charge}.gth'
    else:
        files = glob(f'{psp_path}{system}-q*')
        files.sort()
        try:
            f_psp = files[0]
        except IndexError:
            print(f'ERROR: There is no GTH pseudopotential in {psp_path} for "{system}"')
        if len(files) > 1:
            print(f'INFO: Multiple pseudopotentials found for "{system}". '
                  f'Continue with "{basename(f_psp)}".')

    psp = {}
    C = np.zeros(4)
    rc = np.zeros(4)
    Nproj_l = np.zeros(4, dtype=int)
    h = np.zeros([4, 3, 3])
    try:
        with open(f_psp, 'r') as fh:
            psp['symbol'] = fh.readline().split()[0]
            N_all = fh.readline().split()
            N_s, N_p, N_d, N_f = int(N_all[0]), int(N_all[1]), int(N_all[2]), int(N_all[3])
            psp['Zval'] = N_s + N_p + N_d + N_f
            rlocal, n_c_local = fh.readline().split()
            psp['rlocal'] = float(rlocal)
            psp['n_c_local'] = int(n_c_local)
            for i, val in enumerate(fh.readline().split()):
                C[i] = float(val)
            psp['C'] = C
            lmax = int(fh.readline().split()[0])
            psp['lmax'] = lmax
            for iiter in range(lmax):
                rc[iiter], Nproj_l[iiter] = [float(i) for i in fh.readline().split()]
                for jiter in range(Nproj_l[iiter]):
                    tmp = fh.readline().split()
                    for iread, kiter in enumerate(range(jiter, Nproj_l[iiter])):
                        h[iiter, jiter, kiter] = float(tmp[iread])
            psp['rc'] = rc
            psp['Nproj_l'] = Nproj_l
            for k in range(3):
                for i in range(2):
                    for j in range(i + 1, 2):
                        h[k, j, i] = h[k, i, j]
            psp['h'] = h
    except FileNotFoundError:
        print(f'ERROR: There is no GTH pseudopotential for "{basename(f_psp)}"')
    return psp



#!/usr/bin/env python3
'''
Utilities to localize and analyze orbitals.
'''
import numpy as np
from numpy.linalg import eig, norm


def eval_psi(atoms, psi, r):
    '''Evaluate orbitals at given coordinate points.

    Args:
        atoms :
            Atoms object.

        psi : array
            Set of orbitals in reciprocal space.

        r : array
            Real-space positions.

    Returns:
        Values of psi at points r as an array.
    '''
    # Shift the evaluation point to (0,0,0), because we always have a lattice point there
    psi_T = atoms.T(psi, -r)
    psi_Trs = atoms.I(psi_T)
    # The zero entry is always the value at point (0,0,0)
    return psi_Trs[0]


def get_R(atoms, psi, fods):
    '''Calculate transformation matrix to build Fermi orbitals from Kohn-Sham orbitals.

    Args:
        atoms :
            Atoms object.

        psi : array
            Set of orbitals in reciprocal space.

        fods : array
            Fermi-orbital descriptors.

    Returns:
        Transformation matrix R as an array.
    '''
    # We only calculate occupied orbitals, so a zero matrix is enough
    R = np.empty((len(fods), len(fods)), dtype=complex)
    for i in range(len(fods)):
        # Get the value at one FOD position for all psi
        psi_fod = eval_psi(atoms, psi, fods[i])
        sum_psi_fod = np.sqrt(np.sum(psi_fod.conj() * psi_fod))
        for j in range(len(fods)):
            R[i, j] = psi_fod[j].conj() / sum_psi_fod
    return R


def get_FOs(atoms, psi, fods):
    '''Calculate Fermi orbitals from Kohn-Sham orbitals and a set of respective FODs.

    Args:
        atoms :
            Atoms object.

        psi : array
            Set of orbitals in reciprocal space.

        fods : array
            Fermi-orbital descriptors.

    Returns:
        Fermi orbitals as an array.
    '''
    FOs = np.zeros((len(atoms.r), atoms.Ns), dtype=complex)
    # Get the transformation matrix R
    R = get_R(atoms, psi, fods)
    # Transform psi to real space
    psi_rs = atoms.I(psi)
    for i in range(len(R)):
        for j in range(atoms.Ns):
            FOs[:, i] += R[i, j] * psi_rs[:, j]
    return FOs


def get_S(atoms, psirs):
    '''Calculate overlap matrix between orbitals.

    Args:
        atoms :
            Atoms object.

        psirs : array
            Set of orbitals in real-space.

    Returns:
        Overlap matrix S as an array.
    '''
    # Overlap elements: S_ij = \int psi_i^* psi_j dr
    S = np.empty((atoms.Ns, atoms.Ns), dtype=complex)

    dV = atoms.CellVol / np.prod(atoms.S)
    for i in range(atoms.Ns):
        for j in range(atoms.Ns):
            S[i][j] = dV * np.sum(psirs[:, i].conj() * psirs[:, j])
    return S


def get_FLOs(atoms, psi, fods):
    '''Calculate Fermi-Löwdin orbitals by orthonormalizing Fermi orbitals.

    Args:
        atoms :
            Atoms object.

        psi : array
            Set of orbitals in reciprocal space.

        fods : array
            Fermi-orbital descriptors.

    Returns:
        Fermi-Löwdin orbitals as an array.
    '''
    FOs = get_FOs(atoms, psi, fods)
    # Calculate the overlap matrix for FOs
    S = get_S(atoms, FOs)
    # Calculate eigenvalues and eigenvectors
    Q, T = eig(S)
    # Löwdins symmetric orthonormalization method
    Q12 = np.diag(1 / np.sqrt(Q))
    return FOs @ (T @ Q12 @ T.T)


def wannier_cost(atoms, psirs):
    '''Calculate the Wannier cost function, namely the spread. Equivalent to Foster-Boys.

    Args:
        atoms :
            Atoms object.

        psirs : array
            Set of orbitals in real-space.

    Returns:
        Spread per orbital as an array.
    '''
    centers = wannier_center(atoms, psirs)
    moments = second_moment(atoms, psirs)
    costs = moments - norm(centers, axis=1)**2
    if atoms.verbose >= 3:
        print(f'Centers:\n{centers}\nMoments:\n{moments}')
    if atoms.verbose >= 2:
        print(f'Costs:\n{costs}')
    return np.sum(costs)


def wannier_center(atoms, psirs):
    '''Calculate Wannier centers that are the expecation values of r.

    Args:
        atoms :
            Atoms object.

        psirs : array
            Set of orbitals in real-space.

    Returns:
        Wannier centers per orbital as an array.
    '''
    dV = atoms.CellVol / np.prod(atoms.S)
    r = atoms.r

    centers = np.empty((psirs.shape[1], 3))
    for i in range(psirs.shape[1]):
        for dim in range(3):
            centers[i][dim] = dV * np.real(np.sum(psirs[:, i].conj() * r[:, dim] * psirs[:, i],
                              axis=0))
    return centers


def second_moment(atoms, psirs):
    '''Calculate the second moments that are the expecation values of r^2.

    Args:
        atoms :
            Atoms object.

        psirs : array
            Set of orbitals in real-space.

    Returns:
        Second moments per orbital as an array.
    '''
    dV = atoms.CellVol / np.prod(atoms.S)
    r2 = norm(atoms.r, axis=1)**2

    moments = np.empty(psirs.shape[1])
    for i in range(psirs.shape[1]):
        moments[i] = dV * np.real(np.sum(psirs[:, i].conj() * r2 * psirs[:, i], axis=0))
    return moments



#!/usr/bin/env python3
'''
Package version number and version info function.
'''
from sys import version

__version__ = '1.0.1'


def info():
    '''Print version numbers and availability of used packages.'''
    print('--- Version infos ---')
    print(f'python       : {version.split()[0]}')
    print(f'eminus       : {__version__}')
    for pkg in ('numpy', 'scipy', 'vispy', 'nglview', 'jupyter_rfb', 'pyflosic_dev'):
        try:
            exec(f'import {pkg}')
            print(f'{pkg.ljust(13)}: {eval(pkg).__version__}')
        except ModuleNotFoundError:
            print(f'{pkg.ljust(13)}: not installed')
    return



#!/usr/bin/env python3
'''
Atoms object definitions.
'''
from re import sub

import numpy as np
from numpy.linalg import det, eig, inv
from scipy.fft import next_fast_len

from .energies import Energy
from .gth import init_gth_loc, init_gth_nonloc, read_gth
from .operators import I, Idag, J, Jdag, K, L, Linv, O, T
from .potentials import init_pot
from .tools import center_of_mass, cutoff2gridspacing, inertia_tensor


class Atoms:
    '''Atoms object that holds all necessary calculation parameters.

    Args:
        atom : str or list of str
            Atom symbols.
            Examples: 'CH4'; ['C', 'H', 'H', 'H', 'H']

        X : list or array of floats
            Atom positions.
            Examples: [0, 0, 0]; array([0, 0, 0]); [[0, 0, 0], [1, 1, 1]];
            array([[[0, 0, 0], [1, 1, 1]]])

    Kwargs:
        a : float list or array of floats
            Cell size or vacuum size. A cuboidic box with the same side lengths will be created.
            Default: 20 Bohr (ca. 10.5 Angstrom).

        ecut : float or None
            Cut-off energy. None will disable the G-Vector truncation (needs a separate S).
            Default: 20 Hartree (ca. 544 eV).

        Z : int or list or array of ints or None
            Valence charge per atom. The charges should not differ for same species.
            None will use valence charges from GTH pseudopotentials. The same charge for every atom
            will be assumed for single integers.
            Example: 1; [4, 1, 1, 1, 1]
            Default: None

        S : int or list or array of ints
            Real-space sampling of the cell/vacuum. None will make the sampling dependent of a and
            ecut.
            Example: 30; [30, 40, 50]; array([30, 40, 50])
            Default: None

        f : float or list or array of floats
            Occupation numbers per state. None will assume occupations of 2. The last state will
            be adjusted if the sum of f is not equal to the sum of Z.
            Example: 2; [2, 2/3, 2/3, 2/3]; array([2, 2/3, 2/3, 2/3])
            Default: None

        Ns : int
            Number of states. None will get the number of states from f or assume occupations of 2
            and divide the sum of Z by it.
            Default: None

        verbose : int
            Level of output. Larger numbers mean more output.
            Default: 3

        pot : str
            Type of pseudopotential (case insensitive).
            Example: 'GTH'; 'harmonic'; 'Coulomb'; 'Ge'
            Default: 'gth'

        center : bool
            Center the system inside the box by its geometric center of mass and rotate it such
            that its geometric moment of inertia aligns with the coordinate axes.
            Default: False

        exc : str
            Exchange-correlation functional description (case insensitive), separated by a comma.
            Example: 'lda,vwn'; 'lda,pw'; 'lda,'; ',vwn';0 ',pw'; ','
            Default: 'lda,vwn'

        spinpol : bool
            Spin-polarized calculation.
            Default: False

        cutcoul : float
            Radius of the spherical truncation of the Coulomb potential. None will set the
            theoretical minimum of sqrt(3)*a.
            Default: None
    '''

    def __init__(self, atom, X, a=20, ecut=20, Z=None, S=None, f=None, Ns=None, verbose=3,
                 pot='gth', center=False, exc='lda,vwn', spinpol=False, cutcoul=None):
        self.atom = atom          # Atom symbols
        self.X = X                # Atom positions
        self.a = a                # Cell/Vacuum size
        self.ecut = ecut          # Cut-off energy
        self.Z = Z                # Valence charges
        self.S = S                # Cell sampling
        self.f = f                # Occupation numbers
        self.Ns = Ns              # Number of states
        self.pot = pot            # Used pseudopotential
        self.verbose = verbose    # Output control
        self.center = center      # Center molecule in cell
        self.exc = exc            # Exchange-correlation functional
        self.spinpol = spinpol    # Bool for spin polarized calculations
        self.cutcoul = cutcoul    # Cut-off radius for a spherical coulomb truncation

        # Parameters that will be built out of the inputs
        self.Natoms = None    # Number of atoms
        self.R = None         # Unit cell
        self.CellVol = None   # Unit cell volume
        self.r = None         # Sample points in unit cell
        self.G = None         # G-vectors
        self.G2 = None        # Squared magnitudes of G-vectors
        self.Sf = None        # Structure factor
        self.active = None    # Mask for active G-vectors
        self.Gc = None        # Truncated G-vectors
        self.G2c = None       # Truncated squared magnitudes of G-vectors
        self.GTH = {}         # Dictionary of GTH parameters per atom species
        self.Vloc = None      # Local pseudopotential contribution
        self.NbetaNL = 0      # Number of projector functions for the non-local gth potential
        self.prj2beta = None  # Index matrix to map to the correct projector function
        self.betaNL = None    # Atomic-centered projector functions

        # Parameters after SCF calculations
        self.W = None             # Basis functions
        self.n = None             # Electronic density
        self.energies = Energy()  # Energy object that holds energy contributions

        self.update()

    def update(self):
        '''Check inputs and update them if none are given.'''
        # If a string is given for atom symbols convert them to a list of strings
        if isinstance(self.atom, str):
            # Insert a whitespace before every capital letter, these can appear once or none at all
            # Or insert before digits, these can appear at least once
            self.atom = sub(r'([A-Z?]|\d+)', r' \1', self.atom).split()
            atom = []
            for ia in self.atom:
                if ia.isdigit():
                    # if ia is a integer, append the previous atom ia-1 times
                    atom += [atom[-1]] * (int(ia) - 1)
                else:
                    # If ia is a string, add it to the results list
                    atom += [ia]
            self.atom = atom

        self.Natoms = len(self.atom)

        # We need atom positions as a two-dimensional array
        self.X = np.asarray(self.X)
        if self.X.ndim == 1:
            self.X = np.array([self.X])

        # If only one charge is given, assume it is the charge for every atom
        if isinstance(self.Z, (int, np.integer)):
            self.Z = [self.Z] * self.Natoms
        if isinstance(self.Z, (list, tuple)):
            self.Z = np.asarray(self.Z)

        if isinstance(self.a, (int, np.integer, float, np.floating)):
            self.a = self.a * np.ones(3)
        if isinstance(self.a, (list, tuple)):
            self.a = np.asarray(self.a)

        # Make sampling dependent of ecut if no sampling is given
        if self.S is None:
            try:
                S = np.int_(self.a / cutoff2gridspacing(self.ecut))
            except TypeError:
                print('ERROR: No ecut provided, please enter a valid S.')
            # Multiply by two and add one to match PWDFT
            S = 2 * S + 1
            # Calculate a fast length to optimize the FFT calculations
            # See https://github.com/scipy/scipy/blob/master/scipy/fft/_helper.py
            for i in range(len(S)):
                S[i] = next_fast_len(S[i])
            self.S = S
        # Choose the same sampling for every direction if an integer is given
        if isinstance(self.S, (int, np.integer)):
            self.S = self.S * np.ones(3, dtype=int)
        if isinstance(self.S, (list, tuple)):
            self.S = np.asarray(self.S)

        # Lower the potential string
        self.pot = self.pot.lower()

        # If the cut-off radius is zero, set it to the theoretical minimal value
        if self.cutcoul == 0:
            self.cutcoul = np.sqrt(3) * self.a[0]

        # Center molecule by its geometric center of mass in the unit cell
        # Also rotate it such that the geometric inertia tensor will be diagonal
        if self.center:
            # Rotate the system
            X = self.X
            I = inertia_tensor(X)
            _, eigvecs = eig(I)
            X = np.dot(inv(eigvecs), X.T).T

            # Shift to center of the box
            com = center_of_mass(X)
            self.X = X - (com - self.a / 2)

        # Lower the exchange-correlation string
        self.exc = self.exc.lower()

        # Build a cuboidic unit cell and calculate its volume
        R = self.a * np.eye(3)
        self.R = R
        self.CellVol = np.abs(det(R))

        # Build index matrix M
        ms = np.arange(np.prod(self.S))
        m1 = ms % self.S[0]
        m2 = np.floor(ms / self.S[0]) % self.S[1]
        m3 = np.floor(ms / (self.S[0] * self.S[1])) % self.S[2]
        M = np.array([m1, m2, m3]).T

        # Build index matrix N
        n1 = m1 - (m1 > self.S[0] / 2) * self.S[0]
        n2 = m2 - (m2 > self.S[1] / 2) * self.S[1]
        n3 = m3 - (m3 > self.S[2] / 2) * self.S[2]
        N = np.array([n1, n2, n3]).T

        # Build sampling points
        r = M @ inv(np.diag(self.S)) @ R.T
        self.r = r

        # Build G-vectors
        G = 2 * np.pi * N @ inv(R)
        self.G = G

        # Calculate squared-magnitudes of G-vectors
        G2 = np.sum(G**2, axis=1)
        self.G2 = G2

        # Calculate the structure factor per atom
        Sf = np.exp(1j * G @ self.X.conj().T).T
        self.Sf = Sf

        # Restrict the G and G2
        if self.ecut is not None:
            active = np.nonzero(G2 <= 2 * self.ecut)
        else:
            active = np.nonzero(G2 >= 0)  # Trivial condition to produce the right shape
        self.active = active
        self.Gc = G[active]
        self.G2c = G2[active]

        # Update the potentials
        if self.pot == 'gth':
            if self.Z is not None:
                for ia in range(self.Natoms):
                    self.GTH[self.atom[ia]] = read_gth(self.atom[ia], self.Z[ia])
            else:
                # If no charges are given, use the ones provided by the pseudopotential
                Z = []
                for ia in range(self.Natoms):
                    self.GTH[self.atom[ia]] = read_gth(self.atom[ia])
                    Z.append(self.GTH[self.atom[ia]]['Zval'])
                self.Z = np.asarray(Z)
            # Setup potentials
            self.Vloc = init_gth_loc(self)
            self.NbetaNL, self.prj2beta, self.betaNL = init_gth_nonloc(self)
        else:
            self.Vloc = init_pot(self)

        # Check occupations and number of states together
        if isinstance(self.f, (list, tuple)):
            self.f = np.asarray(self.f)
        # If no states are provided, use the length of the occupations
        if isinstance(self.f, np.ndarray) and self.Ns is None:
            self.Ns = len(self.f)
        # If one occupation and the number of states is given, use it for every state
        if isinstance(self.f, (int, np.integer, float, np.floating)) and self.Ns is not None:
            self.f = self.f * np.ones(self.Ns)
        # If no occupation and the number of states is given, assume 2
        if self.f is None and self.Ns is not None:
            self.f = 2 * np.ones(self.Ns)
        # If number of states is None and occupations is a Number or None, we are in trouble
        if self.Ns is None:
            # If no occupations is given, assume 2
            if self.f is None:
                self.f = 2
            # Assume the number of states by dividing the total valence charge by the occupation
            Ztot = np.sum(self.Z)
            self.Ns = int(np.ceil(Ztot / self.f))
        # If the sum of valence charges is not divisible by occupation, change the last occupation
        if isinstance(self.f, (int, np.integer, float, np.floating)):
            mod = np.sum(self.Z) % self.f
            self.f = self.f * np.ones(self.Ns)
            if mod != 0:
                self.f[-1] = mod
        return

    def __repr__(self):
        atom = self.atom
        Natoms = self.Natoms
        X = self.X
        Z = self.Z

        out = 'Atom\tCharge\tPosition'
        for i in range(Natoms):
            out = f'{out}\n{atom[i]}\t{Z[i]}\t{X[i][0]:10.5f}  {X[i][1]:10.5f}  {X[i][2]:10.5f}'
        return out

    def O(self, inp):
        '''Overlap operator :func:`~eminus.operators.O`.'''
        return O(self, inp)

    def L(self, inp):
        '''Laplacian operator :func:`~eminus.operators.L`.'''
        return L(self, inp)

    def Linv(self, inp):
        '''Inverse Laplacian operator :func:`~eminus.operators.Linv`.'''
        return Linv(self, inp)

    def K(self, inp):
        '''Preconditioning operator :func:`~eminus.operators.K`.'''
        return K(self, inp)

    def I(self, inp):
        '''Transformation from k- to real-space :func:`~eminus.operators.I`.'''
        return I(self, inp)

    def J(self, inp, full=True):
        '''Transformation from real- to k-space :func:`~eminus.operators.J`.'''
        return J(self, inp, full)

    def Idag(self, inp):
        '''Conj transformation from k- to real-space :func:`~eminus.operators.Idag`.'''
        return Idag(self, inp)

    def Jdag(self, inp):
        '''Conj transformation from real- to k-space :func:`~eminus.operators.Jdag`.'''
        return Jdag(self, inp)

    def T(self, inp, dr):
        '''Translation operator :func:`~eminus.operators.T`.'''
        return T(self, inp, dr)



#!/usr/bin/env python3
'''
Collection of constants and unit conversion tools.
'''
# Ha in eV (https://en.wikipedia.org/wiki/Hartree)
electronvolt = eV = 27.211386245988
# Ha in kcal/mol (https://en.wikipedia.org/wiki/Hartree)
kcalmol = 627.5094740631
# a0 in Å (https://en.wikipedia.org/wiki/Bohr_radius)
angstrom = A = 0.529177210903
# e x a0 in D (https://en.wikipedia.org/wiki/Hartree_atomic_units)
Debye = D = 2.541746473


def ha2ev(E):
    '''Convert Hartree to electronvolt.

    Args:
        E : float
            Energy in Hartree.

    Returns:
        Energy in electronvolt as a float.
    '''
    return E * electronvolt


def ev2ha(E):
    '''Convert electronvolt to Hartree.

    Args:
        E : float
            Energy in electronvolt.

    Returns:
        Energy in Hartree as a float.
    '''
    return E / electronvolt


def ha2kcalmol(E):
    '''Convert Hartree to kcal/mol.

    Args:
        E : float
            Energy in Hartree.

    Returns:
        Energy in kcal/mol as a float.
    '''
    return E * kcalmol


def kcalmol2ha(E):
    '''Convert kcal/mol to Hartree.

    Args:
        E : float
            Energy in kcal/mol.

    Returns:
        Energy in Hartree as a float.
    '''
    return E / kcalmol


def ev2kcalmol(E):
    '''Convert electronvolt to kcal/mol.

    Args:
        E : float
            Energy in electronvolt.

    Returns:
        Energy in kcal/mol as a float.
    '''
    return ha2kcalmol(ev2ha(E))


def kcalmol2ev(E):
    '''Convert kcal/mol to electronvolt.

    Args:
        E : float
            Energy in kcal/mol.

    Returns:
        Energy in electronvolt as a float.
    '''
    return ha2ev(kcalmol2ha(E))


def ha2ry(E):
    '''Convert Hartree to Rydberg.

    Args:
        E : float
            Energy in Hartree.

    Returns:
        Energy in Rydberg as a float.
    '''
    return 2 * E


def ry2ha(E):
    '''Convert Rydberg to Hartree.

    Args:
        E : float
            Energy in Rydberg.

    Returns:
        Energy in Hartree as a float.
    '''
    return E / 2


def ang2bohr(r):
    '''Convert Angstrom to Bohr.

    Args:
        r : float
            Length in Angstrom.

    Returns:
        Length in Bohr as a float.
    '''
    return r / angstrom


def bohr2ang(r):
    '''Convert Bohr to Angstrom.

    Args:
        r : float
            Length in Bohr.

    Returns:
        Length in Angstrom as a float.
    '''
    return r * angstrom


def ebohr2d(p):
    '''Convert e x Bohr to Debye.

    Args:
        p : float
            Electric dipole moment in e times Bohr.

    Returns:
        Electric dipole moment in Debye as a float.
    '''
    return p * Debye


def d2ebohr(p):
    '''Convert Debye to e x Bohr.

    Args:
        p : float
            Electric dipole moment in Debye.

    Returns:
        Electric dipole moment in e times Bohr as a float.
    '''
    return p / Debye



#!/usr/bin/env python3
'''
Linear algebra calculation utilities.
'''
import numpy as np
from numpy.linalg import norm


def Diagprod(a, B):
    '''Calculate the expression Diag(a) * B.

    Args:
        a : array
            Single vector.

        B : array
            Matrix.

    Returns:
        Result as an array.
    '''
    return (a * B.T).T


def dotprod(a, b):
    '''Calculate the expression a * b.

    Args:
        a : array
            Array of vectors.

        b : array
            Array of vectors.

    Returns:
        Result as a float.
    '''
    return np.real(np.trace(a.conj().T @ b))


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/Ylm_real.jl
def Ylm_real(l, m, R):
    '''Calculate spherical harmonics.

    Args:
        l : int
            Angular momentum number.

        m : int
            Magnetic quantum number.

        R : array
            Real-space position vector or array of real-space position vectors.

    Returns:
        Spherical harmonics as an array.
    '''
    # Account for single vectors
    if R.ndim == 1:
        R = np.array([R])

    # No need to calculate more for l=0
    if l == 0:
        return 0.5 * np.sqrt(1 / np.pi) * np.ones(len(R))

    # Account for small norm, if norm(R) < eps: cost=0
    eps = 1e-9
    Rmod = norm(R, axis=1)
    cost = np.zeros(len(R))
    cost_idx = Rmod > eps
    cost[cost_idx] = R[cost_idx, 2] / Rmod[cost_idx]

    # Vectorized version of sqrt(max(0, 1-cost^2))
    sint = np.sqrt(np.amax([np.zeros(len(R)), 1 - cost**2], axis=0))

    # If Rx=0: phi=pi/2*sign(Ry)
    phi = -np.pi / 2 * np.ones(len(R))
    phi_idx = R[:, 1] >= 0
    phi[phi_idx] = np.pi / 2 * np.ones(len(phi[phi_idx]))

    # Beware the arctan, it is defined with modulo pi
    phi_idx = R[:, 0] < -eps
    phi[phi_idx] = np.arctan(R[phi_idx, 1] / R[phi_idx, 0]) + np.pi
    phi_idx = R[:, 0] > eps
    phi[phi_idx] = np.arctan(R[phi_idx, 1] / R[phi_idx, 0])

    if l == 1:
        if m == -1:   # py
            ylm = 0.5 * np.sqrt(3 / np.pi) * sint * np.sin(phi)
        elif m == 0:  # pz
            ylm = 0.5 * np.sqrt(3 / np.pi) * cost
        elif m == 1:  # px
            ylm = 0.5 * np.sqrt(3 / np.pi) * sint * np.cos(phi)
        else:
            print(f'ERROR: No definition found for Ylm({l}, {m})')
    elif l == 2:
        if m == -2:    # dxy
            ylm = np.sqrt(15 / 16 / np.pi) * sint**2 * np.sin(2 * phi)
        elif m == -1:  # dyz
            ylm = np.sqrt(15 / 4 / np.pi) * cost * sint * np.sin(phi)
        elif m == 0:   # dz2
            ylm = 0.25 * np.sqrt(5 / np.pi) * (3 * cost**2 - 1)
        elif m == 1:   # dxz
            ylm = np.sqrt(15 / 4 / np.pi) * cost * sint * np.cos(phi)
        elif m == 2:   # dx2-y2
            ylm = np.sqrt(15 / 16 / np.pi) * sint**2 * np.cos(2 * phi)
        else:
            print(f'ERROR: No definition found for Ylm({l}, {m})')
    elif l == 3:
        if m == -3:
            ylm = 0.25 * np.sqrt(35 / 2 / np.pi) * sint**3 * np.sin(3 * phi)
        elif m == -2:
            ylm = 0.25 * np.sqrt(105 / np.pi) * sint**2 * cost * np.sin(2 * phi)
        elif m == -1:
            ylm = 0.25 * np.sqrt(21 / 2 / np.pi) * sint * (5 * cost**2 - 1) * np.sin(phi)
        elif m == 0:
            ylm = 0.25 * np.sqrt(7 / np.pi) * (5 * cost**3 - 3 * cost)
        elif m == 1:
            ylm = 0.25 * np.sqrt(21 / 2 / np.pi) * sint * (5 * cost**2 - 1) * np.cos(phi)
        elif m == 2:
            ylm = 0.25 * np.sqrt(105 / np.pi) * sint**2 * cost * np.cos(2 * phi)
        elif m == 3:
            ylm = 0.25 * np.sqrt(35 / 2 / np.pi) * sint**3 * np.cos(3 * phi)
        else:
            print(f'ERROR: No definition found for Ylm({l}, {m})')
    else:
        print(f'ERROR: No definition found for Ylm({l}, {m})')
    return ylm



#!/usr/bin/env python3
'''
Basis set depedent operators for a plane-wave basis.
'''
from os import environ

import numpy as np
from scipy.fft import ifftn, fftn

try:
    THREADS = int(environ['OMP_NUM_THREADS'])
except KeyError:
    THREADS = None


def O(atoms, inp):
    '''Overlap operator.

    Args:
        atoms :
            Atoms object.

        inp : array
            Coefficents input array.

    Returns:
        Result as an array.
    '''
    return atoms.CellVol * inp


def L(atoms, inp):
    '''Laplacian operator.

    Args:
        atoms :
            Atoms object.

        inp : array
            Coefficents input array.

    Returns:
        Result as an array.
    '''
    inp = inp.T
    if inp.shape[1] == len(atoms.G2c):
        return (-atoms.CellVol * atoms.G2c * inp).T
    else:
        return (-atoms.CellVol * atoms.G2 * inp).T


def Linv(atoms, inp):
    '''Inverse Laplacian operator.

    Args:
        atoms :
            Atoms object.

        inp : array
            Coefficents input array.

    Returns:
        Result as an array.
    '''
    inp = inp.T
    out = np.zeros_like(inp, dtype=complex)
    out[0] = 0
    if inp.ndim == 1:
        out[1:] = inp[1:] / atoms.G2[1:] / -atoms.CellVol
    else:
        for i in range(len(inp)):
            out[i][1:] = inp[i][1:] / atoms.G2[1:] / -atoms.CellVol
    return out.T


def K(atoms, inp):
    '''Preconditioning operator. Applies 1/(1+G2) to the input.

    Args:
        atoms :
            Atoms object.

        inp : array
            Coefficents input array.

    Returns:
        Result as an array.
    '''
    inp = inp.T
    out = np.empty_like(inp, dtype=complex)
    if inp.shape[1] == len(atoms.G2c):
        if inp.ndim == 1:
            out = inp / (1 + atoms.G2c)
        else:
            for i in range(len(inp)):
                out[i] = inp[i] / (1 + atoms.G2c)
    else:
        if inp.ndim == 1:
            out = inp / (1 + atoms.G2)
        else:
            for i in range(len(inp)):
                out[i] = inp[i] / (1 + atoms.G2)
    return out.T


def I(atoms, inp):
    '''Backwards transformation from reciprocal space to real-space.

    Args:
        atoms :
            Atoms object.

        inp : array
            Coefficents input array.

    Returns:
        Result as an array.
    '''
    inp = inp.T
    if inp.ndim == 1:
        inp = np.array([inp])
    if np.size(inp, 1) == np.prod(atoms.S):
        out = np.empty_like(inp, dtype=complex)
        for i in range(inp.shape[0]):
            tmp = np.reshape(inp[i], atoms.S, order='F')
            out[i] = ifftn(tmp, workers=THREADS).flatten(order='F')
    else:
        out = np.empty((inp.shape[0], np.prod(atoms.S)), dtype=complex)
        for i in range(inp.shape[0]):
            full = np.zeros(np.prod(atoms.S), dtype=complex)
            full[atoms.active] = inp[i]
            full = np.reshape(full, atoms.S, order='F')
            out[i] = ifftn(full, workers=THREADS).flatten(order='F')
    return (out * np.prod(atoms.S)).T


def J(atoms, inp, full=True):
    '''Forward transformation from real-space to reciprocal space.

    Args:
        atoms :
            Atoms object.

        inp : array
            Coefficents input array.

    Kwargs:
        full : bool
            Transform in the full or in the active/truncated space.

    Returns:
        Result as an array.
    '''
    inp = inp.T
    if inp.ndim == 1:
        tmp = np.reshape(inp, atoms.S, order='F')
        out = fftn(tmp, workers=THREADS).flatten(order='F')
        if not full:
            out = out[atoms.active]
    else:
        if full:
            out = np.empty_like(inp, dtype=complex)
            for i in range(inp.shape[0]):
                tmp = np.reshape(inp[i], atoms.S, order='F')
                out[i] = fftn(tmp, workers=THREADS).flatten(order='F')
        else:
            out = np.empty((inp.shape[0], len(atoms.active[0])), dtype=complex)
            for i in range(inp.shape[0]):
                tmp = np.reshape(inp[i], atoms.S, order='F')
                out[i] = fftn(tmp, workers=THREADS).flatten(order='F')[atoms.active]
    return (out / np.prod(atoms.S)).T


def Idag(atoms, inp):
    '''Conjugated backwards transformation from reciprocal space to real-space.

    Args:
        atoms :
            Atoms object.

        inp : array
            Coefficents input array.

    Returns:
        Result as an array.
    '''
    inp = inp.T
    if inp.ndim == 1:
        tmp = np.reshape(inp, atoms.S, order='F')
        full = fftn(tmp, workers=THREADS).flatten(order='F')
        out = full[atoms.active]
    else:
        out = np.empty((np.size(inp, 0), len(atoms.active[0])), dtype=complex)
        for i in range(len(inp)):
            tmp = np.reshape(inp[i], atoms.S, order='F')
            full = fftn(tmp, workers=THREADS).flatten(order='F')
            out[i] = full[atoms.active]
    return out.T


def Jdag(atoms, inp):
    '''Conjugated forward transformation from real-space to reciprocal space.

    Args:
        atoms :
            Atoms object.

        inp : array
            Coefficents input array.

    Returns:
        Result as an array.
    '''
    inp = inp.T
    if inp.ndim == 1:
        tmp = np.reshape(inp, atoms.S, order='F')
        out = ifftn(tmp, workers=THREADS).flatten(order='F')
    else:
        out = np.empty_like(inp, dtype=complex)
        for i in range(len(inp)):
            tmp = np.reshape(inp[i], atoms.S, order='F')
            out[i] = ifftn(tmp, workers=THREADS).flatten(order='F')
    return out.T


def T(atoms, inp, dr):
    '''Translation operator. Shifts input by the vector dr.

    Args:
        atoms :
            Atoms object.

        inp : array
            Coefficents input array.

        dr : array
            Real-space position vector.

    Returns:
        Result as an array.
    '''
    out = np.empty_like(inp, dtype=complex)
    factor = np.exp(-1j * np.dot(atoms.Gc, dr))
    for i in range(atoms.Ns):
        out[:, i] = factor * inp[:, i]
    return out



#!/usr/bin/env python3
from setuptools import find_packages, setup

with open('eminus/version.py', 'r') as fh:
    version = {}
    exec(fh.read(), version)

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='eminus',
    version=version['__version__'],
    description='A plane wave density funtional theory code.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/wangenau/eminus',
    author='Wanja Timm Schulze',
    author_email='wangenau@protonmail.com',
    license='APACHE2.0',
    packages=find_packages(),
    install_requires=['numpy', 'scipy'],
    extras_require={
        'addons': ['jupyter_rfb', 'nglview', 'vispy']  # , 'pyflosic_dev']
    },
    python_requires='>=3.6',
    package_data={'eminus.pade_gth': ['*.gth']},
    zip_safe=False
)



#!/usr/bin/env python3
'''
Parametizations of density functionals.
'''
import numpy as np


def get_exc(exc, n, spinpol):
    '''Handle and get exchange-correlation functionals.

    Args:
        exc : str
            Exchange and correlation identifier, separated by a comma.

        n : array
            Real-space electronic density.

        spinpol : bool
            Choose if a spin-polarized exchange-correlation functional will be used.

    Returns:
        Exchange-correlation energy density and potential as a tuple(array, array).
    '''
    exc_map = {
        'lda': 'lda_slater_x',
        'pw': 'lda_pw_c',
        'vwn': 'lda_vwn_c'
    }

    exch, corr = exc.split(',')
    f_exch = exc_map.get(exch, dummy_exc)
    f_corr = exc_map.get(corr, dummy_exc)
    if spinpol:
        f_exch = f'{f_exch}_spin'
        f_corr = f'{f_corr}_spin'

    # FIXME: In spin-polarized calculations zeta is normally not one, only when coming from
    #        spin-unpolarised calculations
    ex, vx = eval(f_exch)(n, zeta=np.ones_like(n))
    ec, vc = eval(f_corr)(n, zeta=np.ones_like(n))
    return ex + ec, vx + vc


def dummy_exc(n, **kwargs):
    '''Dummy exchange-correlation functional with no effect.

    Args:
        n : array
            Real-space electronic density.

    Returns:
        Dummy exchange-correlation energy density and potential as a tuple(array, array).
    '''
    zero = np.zeros_like(n)
    return zero, zero


dummy_exc_spin = dummy_exc


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_x_slater.jl
def lda_slater_x(n, alpha=2 / 3, **kwargs):
    '''Slater exchange functional (spin paired).

    Args:
        n : array
            Real-space electronic density.

    Kwargs:
        alpha : float
            Scaling factor.

    Returns:
        Exchange energy density and potential as a tuple(array, array).
    '''
    third = 1 / 3
    p43 = 4 / 3
    pi34 = (3 / (4 * np.pi))**third
    f = -9 / 8 * (3 / (2 * np.pi))**(2 / 3)
    rs = pi34 / n**third

    ex = f * alpha / rs
    vx = p43 * f * alpha / rs
    return ex, vx


# Adapted from
# https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_x_slater_spin.jl
def lda_slater_x_spin(n, zeta, alpha=2 / 3):
    '''Slater exchange functional (spin polarized).

    Args:
        n : array
            Real-space electronic density.

        zeta : array
            Relative spin polarization.

    Kwargs:
        alpha : float
            Scaling factor.

    Returns:
        Exchange energy density and potential as a tuple(array, array).
    '''
    third = 1 / 3
    p43 = 4 / 3
    f = -9 / 8 * (3 / np.pi)**third

    rho13 = ((1 + zeta) * n)**third
    exup = f * alpha * rho13
    vxup = p43 * f * alpha * rho13

    rho13 = ((1 - zeta) * n)**third
    exdw = f * alpha * rho13
    vxdw = p43 * f * alpha * rho13
    ex = 0.5 * ((1 + zeta) * exup + (1 - zeta) * exdw)
    return ex, [vxup, vxdw]


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_c_pw.jl
def lda_pw_c(n, **kwargs):
    '''PW parameterization of the correlation functional (spin paired).

    Args:
        n : array
            Real-space electronic density.

    Returns:
        Correlation energy density and potential as a tuple(array, array).
    '''
    third = 1 / 3
    pi34 = (3 / (4 * np.pi))**third
    rs = pi34 / n**third

    a = 0.031091
    a1 = 0.2137
    b1 = 7.5957
    b2 = 3.5876
    b3 = 1.6382
    b4 = 0.49294

    rs12 = np.sqrt(rs)
    rs32 = rs * rs12
    rs2 = rs**2

    om = 2 * a * (b1 * rs12 + b2 * rs + b3 * rs32 + b4 * rs2)
    dom = 2 * a * (0.5 * b1 * rs12 + b2 * rs + 1.5 * b3 * rs32 + 2 * b4 * rs2)
    olog = np.log(1 + 1 / om)
    ec = -2 * a * (1 + a1 * rs) * olog
    vc = -2 * a * (1 + 2 / 3 * a1 * rs) * olog - 2 / 3 * a * (1 + a1 * rs) * dom / (om * (om + 1))
    return ec, vc


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_c_pw_spin.jl
def lda_pw_c_spin(n, zeta, **kwargs):
    '''PW parameterization of the correlation functional (spin polarized).

    Args:
        n : array
            Real-space electronic density.

        zeta : array
            Relative spin polarization.

    Returns:
        Correlation energy density and potential as a tuple(array, array).
    '''
    third = 1 / 3
    pi34 = (3 / (4 * np.pi))**third
    rs = pi34 / n**third

    fz0 = 1.709921

    zeta2 = zeta * zeta
    zeta3 = zeta2 * zeta
    zeta4 = zeta3 * zeta
    rs12 = np.sqrt(rs)
    rs32 = rs * rs12
    rs2 = rs**2

    def pw_fit(i):
        '''Calculate correlation energies by Perdew-Wang approximation interpolation.

        Args:
            i : int
                Index weather to use unpolarised (0), polarised (1), or antiferromagnetic (2) fit.

        Returns:
            PW fit and its derivative as a tuple(array, array).
        '''
        a = (0.031091, 0.015545, 0.016887)
        a1 = (0.2137, 0.20548, 0.11125)
        b1 = (7.5957, 14.1189, 10.357)
        b2 = (3.5876, 6.1977, 3.6231)
        b3 = (1.6382, 3.3662, 0.88026)
        b4 = (0.49294, 0.62517, 0.49671)

        om = 2 * a[i] * (b1[i] * rs12 + b2[i] * rs + b3[i] * rs32 + b4[i] * rs2)
        dom = 2 * a[i] * (0.5 * b1[i] * rs12 + b2[i] * rs + 1.5 * b3[i] * rs32 + 2 * b4[i] * rs2)
        olog = np.log(1 + 1 / om)

        fit = -2 * a[i] * (1 + a1[i] * rs) * olog
        dfit = -2 * a[i] * (1 + 2 / 3 * a1[i] * rs) * olog - \
               2 / 3 * a[i] * (1 + a1[i] * rs) * dom / (om * (om + 1))
        return fit, dfit

    ecU, vcU = pw_fit(0)  # Unpolarised
    ecP, vcP = pw_fit(1)  # Polarized
    ac, dac = pw_fit(2)  # Spin stiffness

    fz = ((1 + zeta)**(4 / 3) + (1 - zeta)**(4 / 3) - 2) / (2**(4 / 3) - 2)
    dfz = ((1 + zeta)**third - (1 - zeta)**third) * 4 / (3 * (2**(4 / 3) - 2))

    ec = ecU + ac * fz * (1 - zeta4) / fz0 + (ecP - ecU) * fz * zeta4

    vcup = vcU + dac * fz * (1 - zeta4) / fz0 + (vcP - vcU) * fz * zeta4 + \
           (ac / fz0 * (dfz * (1 - zeta4) - 4 * fz * zeta3) +
           (ecP - ecU) * (dfz * zeta4 + 4 * fz * zeta3)) * (1 - zeta)

    vcdw = vcU + dac * fz * (1 - zeta4) / fz0 + (vcP - vcU) * fz * zeta4 - \
           (ac / fz0 * (dfz * (1 - zeta4) - 4 * fz * zeta3) +
           (ecP - ecU) * (dfz * zeta4 + 4 * fz * zeta3)) * (1 + zeta)
    return ec, [vcup, vcdw]


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_c_vwn.jl
def lda_vwn_c(n, **kwargs):
    '''VWN parameterization of the correlation functional (spin paired).

    Args:
        n : array
            Real-space electronic density.

    Returns:
        Correlation energy density and potential as a tuple(array, array).
    '''
    third = 1 / 3
    pi34 = (3 / (4 * np.pi))**third
    rs = pi34 / n**third

    A = 0.0310907
    b = 3.72744
    c = 12.9352
    x0 = -0.10498

    q = np.sqrt(4 * c - b * b)
    f1 = 2 * b / q
    f2 = b * x0 / (x0 * x0 + b * x0 + c)
    f3 = 2 * (2 * x0 + b) / q
    rs12 = np.sqrt(rs)
    fx = rs + b * rs12 + c
    qx = np.arctan(q / (2 * rs12 + b))
    ec = A * (np.log(rs / fx) + f1 * qx - f2 * (np.log((rs12 - x0)**2 / fx) + f3 * qx))

    tx = 2 * rs12 + b
    tt = tx * tx + q * q
    vc = ec - rs12 * A / 6 * (2 / rs12 - tx / fx - 4 * b / tt -
         f2 * (2 / (rs12 - x0) - tx / fx - 4 * (2 * x0 + b) / tt))
    return ec, vc


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_c_vwn_spin.jl
def lda_vwn_c_spin(n, zeta, **kwargs):
    '''VWN parameterization of the correlation functional (spin polarized).

    Args:
        n : array
            Real-space electronic density.

        zeta : array
            Relative spin polarization.

    Returns:
        Correlation energy density and potential as a tuple(array, array).
    '''
    third = 1 / 3
    pi34 = (3 / (4 * np.pi))**third
    rs = pi34 / n**third

    cfz = 2**(4 / 3) - 2
    cfz1 = 1 / cfz
    cfz2 = 4 / 3 * cfz1
    iddfz0 = 9 / 8 * cfz
    sqrtrs = np.sqrt(rs)
    zeta3 = zeta**3
    zeta4 = zeta3 * zeta
    trup = 1 + zeta
    trdw = 1 - zeta
    trup13 = trup**third
    trdw13 = trdw**third
    fz = cfz1 * (trup13 * trup + trdw13 * trdw - 2)  # f(zeta)
    dfz = cfz2 * (trup13 - trdw13)  # df/dzeta

    def pade_fit(i):
        '''Calculate correlation energies by Pade approximation interpolation.

        Args:
            i : int
                Index weather to use paramagnetic (0), ferromagnetic (1), or spin stiffness (2) fit.

        Returns:
            Pade fit and its derivative as a tuple(array, array).
        '''
        A = (0.0310907, 0.01554535, -0.01688686394039)
        b = (3.72744, 7.06042, 1.13107)
        c = (12.9352, 18.0578, 13.0045)
        x0 = (-0.10498, -0.325, -0.0047584)
        Q = (6.15199081975908, 4.73092690956011, 7.12310891781812)
        tbQ = (1.21178334272806, 2.98479352354082, 0.31757762321188)
        bx0fx0 = (-0.03116760867894, -0.14460061018521, -0.00041403379428)

        xx0 = sqrtrs - x0[i]
        Qtxb = Q[i] / (2 * sqrtrs + b[i])
        atg = np.arctan(Qtxb)
        fx = rs + b[i] * sqrtrs + c[i]

        fit = A[i] * (np.log(rs / fx) + tbQ[i] * atg -
              bx0fx0[i] * (np.log(xx0 * xx0 / fx) + (tbQ[i] + 4 * x0[i] / Q[i]) * atg))

        txb = 2 * sqrtrs + b[i]
        txbfx = txb / fx
        itxbQ = 1 / (txb * txb + Q[i] * Q[i])

        dfit = fit - A[i] / 3 + A[i] * sqrtrs / 6 * (txbfx + 4 * b[i] * itxbQ +
               bx0fx0[i] * (2 / xx0 - txbfx - 4 * (b[i] + 2 * x0[i]) * itxbQ))
        return fit, dfit

    ecP, vcP = pade_fit(0)  # Paramagnetic fit
    ecF, vcF = pade_fit(1)  # Ferromagnetic fit
    ac, dac = pade_fit(2)  # Spin stiffness

    ac = ac * iddfz0
    dac = dac * iddfz0
    De = ecF - ecP - ac  # e_c[F] - e_c[P] - alpha_c/(ddf/ddz(z=0))
    fzz4 = fz * zeta4
    ec = ecP + ac * fz + De * fzz4

    dec1 = vcP + dac * fz + (vcF - vcP - dac) * fzz4  # e_c-(r_s/3)*(de_c/dr_s)
    dec2 = ac * dfz + De * (4 * zeta3 * fz + zeta4 * dfz)  # de_c/dzeta

    vcup = dec1 + (1 - zeta) * dec2
    vcdw = dec1 - (1 + zeta) * dec2
    return ec, [vcup, vcdw]



#!/usr/bin/env python3
'''
Import and export functionalities.
'''
from pickle import dump, HIGHEST_PROTOCOL, load
from textwrap import fill
from time import ctime

import numpy as np

from .data import number2symbol, symbol2number
from .units import ang2bohr, bohr2ang
from .version import __version__


def read_xyz(filename, info=False):
    '''Load atom species and positions from xyz files.

    Args:
        filename : str
            xyz input file path/name.

    Kwargs:
        info : bool
            Display file comments.

    Returns:
        Atom species and positions as a tuple(list, array).
    '''
    # XYZ file definitions: https://en.wikipedia.org/wiki/XYZ_file_format
    with open(filename, 'r') as fh:
        lines = fh.readlines()

    # The first line contains the number of atoms
    Natoms = int(lines[0].strip())

    # The second line can contain a comment, print it if available
    comment = lines[1].strip()
    if info:
        print(f'XYZ file comment: "{comment}"')

    atom = []
    X = []
    # Following lines contain atom positions with the format: Atom x-pos y-pos z-pos
    for line in lines[2:2 + Natoms]:
        line_split = line.strip().split()
        atom.append(line_split[0])
        X.append(np.float_(line_split[1:4]))

    # xyz files are in angstrom, so convert to bohr
    X = ang2bohr(np.asarray(X))
    return atom, X


def write_xyz(atoms, filename, extra=None):
    '''Generate xyz files from atoms objects.

    Args:
        atoms :
            Atoms object.

        filename : str
            xyz output file path/name.

    Kwargs:
        extra : array
            Extra coordinates to write.
    '''
    # XYZ file definitions: https://en.wikipedia.org/wiki/XYZ_file_format
    atom = atoms.atom
    Natoms = atoms.Natoms
    X = atoms.X

    if not filename.endswith('.xyz'):
        filename = f'{filename}.xyz'

    # Convert the coordinates from atomic units to angstrom
    X = bohr2ang(X)
    if extra is not None:
        extra = bohr2ang(extra)

    with open(filename, 'w') as fp:
        # The first line contains the number of atoms.
        # If we add extra coordinates, add them to the count.
        if extra is None:
            fp.write(f'{Natoms}\n')
        else:
            fp.write(f'{Natoms + len(extra)}\n')
        # The second line can contains a comment.
        # Print informations about the file and program, and the file creation time.
        fp.write(f'XYZ file generated with eminus {__version__} at {ctime()}\n')
        for ia in range(Natoms):
            fp.write(f'{atom[ia]}  {X[ia][0]:.5f}  {X[ia][1]:.5f}  {X[ia][2]:.5f}\n')
        # Add extra coordinates, if desired. The will default to X (no atom type).
        if extra is not None:
            for ie in extra:
                fp.write(f'X  {ie[0]:.5f}  {ie[1]:.5f}  {ie[2]:.5f}\n')
    return


def read_cube(filename, info=False):
    '''Load atom and cell data from cube files.

    Args:
        filename : str
            cube input file path/name.
    Kwargs:
        info : bool
            Display file comments.

    Returns:
        Species, positions, charges, cell size, and sampling as a
        tuple(list, array, float, array, int).
    '''
    # It seems, that there is no standard for cube files. The following definition is taken from:
    # https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html
    # Atomic units and a cuboidic unit cell that starts at (0,0,0) is assumed.
    with open(filename, 'r') as fh:
        lines = fh.readlines()

    # The first and second line can contain comments, print them if available
    comment = f'{lines[0].strip()}\n{lines[1].strip()}'
    if info:
        print(f'XYZ file comment: "{comment}"')

    # Line 4 to 6 contain the sampling per axis, and the unit cell basis vectors with length a/S
    # A cuboidic unit cell is assumed, so only use the diagonal entries
    S = np.empty(3)
    a = np.empty(3)
    for i, line in enumerate(lines[3:6]):
        line_split = line.strip().split()
        S[i] = int(line_split[0])
        a[i] = S[i] * np.float_(line_split[i + 1])

    atom = []
    X = []
    Z = []
    # Following lines contain atom positions with the format: atom-id charge x-pos y-pos z-pos
    for line in lines[6:]:
        line_split = line.strip().split()
        # If the first value is not a (positive) integer, we have reached the field data
        if not line_split[0].isdigit():
            break
        atom.append(number2symbol[int(line_split[0])])
        Z.append(line_split[1])
        X.append(np.float_(line_split[2:5]))

    X = np.asarray(X)
    return atom, X, Z, a, S


def write_cube(atoms, field, filename, extra=None):
    '''Generate cube files from given field data.

    Args:
        atoms :
            Atoms object.

        field : array
            Real-space field data.

        filename : str
            xyz output file path/name.

    Kwargs:
        extra : array
            Extra coordinates to write.
    '''
    # It seems, that there is no standard for cube files. The following definition will work with
    # VESTA and is taken from: https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html
    # Atomic units are assumed, so there is no need for conversion.
    atom = atoms.atom
    Natoms = atoms.Natoms
    a = atoms.a
    r = atoms.r
    S = atoms.S
    X = atoms.X
    Z = atoms.Z

    if not filename.endswith('.cube'):
        filename = f'{filename}.cube'

    # Our field data has been created in a different order than needed for cube files
    # (triple loop over z,y,x instead of x,y,z), so rearrange it with some index magic.
    idx = []
    for Nx in range(S[0]):
        for Ny in range(S[1]):
            for Nz in range(S[2]):
                idx.append(Nx + Ny * S[0] + Nz * S[0] * S[1])
    idx = np.asarray(idx)

    # Make sure we have real valued data in the correct order
    field = np.real(field[idx])

    with open(filename, 'w') as fp:
        # The first two lines have to be a comment.
        # Print file creation time and informations about the file and program.
        fp.write(f'{ctime()}\n')
        fp.write(f'Cube file generated with eminus {__version__}\n')
        # Number of atoms (int), and origin of the coordinate system (float)
        # The origin is normally at 0,0,0 but we could move our box, so take the minimum
        if extra is None:
            fp.write(f'{Natoms}  ')
        else:
            fp.write(f'{Natoms + len(extra)}  ')
        fp.write(f'{min(r[:, 0]):.5f}  {min(r[:, 1]):.5f}  {min(r[:, 2]):.5f}\n')
        # Number of points per axis (int), and vector defining the axis (float)
        # We only have a cuboidic box, so each vector only has one non-zero component
        fp.write(f'{S[0]}  {a[0] / S[0]:.5f}  0.0  0.0\n')
        fp.write(f'{S[1]}  0.0  {a[1] / S[1]:.5f}  0.0\n')
        fp.write(f'{S[2]}  0.0  0.0  {a[2] / S[2]:.5f}\n')
        # Atomic number (int), atomic charge (float), and atom position (floats) for every atom
        for ia in range(Natoms):
            fp.write(f'{symbol2number[atom[ia]]}  {Z[ia]:.5f}  ')
            fp.write(f'{X[ia][0]:.5f}  {X[ia][1]:.5f}  {X[ia][2]:.5f}\n')
        if extra is not None:
            for ie in extra:
                fp.write('0  0.00000  ')
                fp.write(f'{ie[0]:.5f}  {ie[1]:.5f}  {ie[2]:.5f}\n')
        # Field data (float) with scientific formatting
        # We have S[0]*S[1] chunks values with a length of S[2]
        for i in range(S[0] * S[1]):
            # Print every round of values, so we can add empty lines between them
            data_str = '%+1.5e  ' * S[2] % tuple(field[i * S[2]:(i + 1) * S[2]])
            # Print a maximum of 6 values per row
            # Max width for this formatting is 90, since 6*len('+1.00000e-000  ')=90
            fp.write(f'{fill(data_str, width=90)}\n\n')
    return


def save_atoms(atoms, filename):
    '''Save atoms objects to a pickle file.

    Args:
        atoms :
            Atoms object.

        filename : str
            xyz input file path/name.
    '''
    with open(filename, 'wb') as fp:
        dump(atoms, fp, HIGHEST_PROTOCOL)
    return


def load_atoms(filename):
    '''Load atoms objects from a pickle file.

    Args:
        filename : str
            xyz input file path/name.

    Returns:
        Atoms object.
    '''
    with open(filename, 'rb') as fh:
        return load(fh)


def create_pdb(atom, X, a=None):
    '''Convert xyz files to pdb format.

    Args:
        atom : list
            Atom symbols.

        X : array
            Atom positions.

    Kwargs:
        a : float
            Cell size.

    Returns:
        pdb file format as a string.
    '''
    # pdb file definitions:
    # For CRYST1: https://www.wwpdb.org/documentation/file-format-content/format33/sect8.html#CRYST1
    # For ATOM: https://www.wwpdb.org/documentation/file-format-content/format33/sect9.html#ATOM
    # Convert Bohr to Angstrom
    X = bohr2ang(X)
    if a is not None:
        a = bohr2ang(a)

    # pdb files have specific numbers of characters for every data with changing justification
    # Write everything explicitly down to not loose track of line lengths
    pdb = ''
    # Create data for a cuboidic cell
    if a is not None:
        pdb += 'CRYST1'               # 1-6 "CRYST1"
        pdb += f'{a[0]:9.3f}'.rjust(9)   # 7-15 a
        pdb += f'{a[1]:9.3f}'.rjust(9)   # 16-24 b
        pdb += f'{a[2]:9.3f}'.rjust(9)   # 25-33 c
        pdb += f'{90:7.2f}'.rjust(7)  # 34-40 alpha
        pdb += f'{90:7.2f}'.rjust(7)  # 41-47 beta
        pdb += f'{90:7.2f}'.rjust(7)  # 48-54 gamma
        pdb += ' '
        pdb += 'P 1'.ljust(11)        # 56-66 Space group
        # pdb += '1'.rjust(4)         # 67-70 Z value

    # Create molecule data
    pdb += '\nMODEL 1'
    for ia in range(len(atom)):
        pdb += '\nATOM  '                   # 1-6 "ATOM"
        pdb += f'{ia + 1}'.rjust(5)         # 7-11 Atom serial number
        pdb += ' '
        pdb += f'{atom[ia]}'.rjust(4)       # 13-16 Atom name
        pdb += ' '                          # 17 Alternate location indicator
        pdb += 'MOL'                        # 18-20 Residue name
        pdb += ' '
        pdb += ' '                          # 22 Chain identifier
        pdb += '1'.rjust(4)                 # 23-26 Residue sequence number
        pdb += ' '                          # 27 Code for insertions of residues
        pdb += '   '
        pdb += f'{X[ia][0]:8.3f}'.rjust(8)  # 31-38 X orthogonal coordinate
        pdb += f'{X[ia][1]:8.3f}'.rjust(8)  # 39-46 Y orthogonal coordinate
        pdb += f'{X[ia][2]:8.3f}'.rjust(8)  # 47-54 Z orthogonal coordinate
        pdb += f'{1:6.2f}'.rjust(6)         # 55-60 Occupancy
        pdb += f'{0:6.2f}'.rjust(6)         # 61-66 Temperature factor
        pdb += '          '
        pdb += f'{atom[ia]}'.rjust(2)       # 77-78 Element symbol
        # pdb += '  '                       # 79-80 Charge
    pdb = f'{pdb}\nENDMDL'
    return pdb



#!/usr/bin/env python3
'''
Fermi-orbital descriptor generation for localization of orbitals.
'''
from os import remove

import numpy as np
from numpy.linalg import norm
try:
    from pyscf.gto import M  # PySCF is a dependency of PyFLOSIC
    from pyscf.scf import RKS
    from pyflosic_dev.atoms.atoms import Atoms
    from pyflosic_dev.guess.pycom import pycom
    from pyflosic_dev.parameters.flosic_parameters import parameters
except ImportError:
    print('ERROR: Necessary addon dependecies not found. '
          'To use this module, install the package with addons, e.g., with '
          '"pip install eminus[addons]"')

from eminus.atoms_io import read_xyz
from eminus.data import covalent_radii, symbol2number
from eminus.units import bohr2ang


def get_fods(atoms, basis='pc-0', loc='FB', clean=True):
    '''Generate FOD positions using the pycom method.

    Args:
        atoms :
            Atoms object.

    Kwargs:
        basis : str
            Basis set for the DFT calculation.

        loc : str
            Localization method.

        clean : bool
            Remove pycom log files.

    Returns:
        FOD positions as an array.
    '''
    # Convert to Angstrom for PySCF
    X = bohr2ang(atoms.X)
    # Build the PySCF input format
    atom_pyscf = [i for i in zip(atoms.atom, X)]

    # Do the PySCF DFT calculation
    spin = np.sum(atoms.Z) % 2
    mol = M(atom=atom_pyscf, basis=basis, spin=spin)
    mf = RKS(mol=mol)
    mf.verbose = 0
    mf.kernel()

    # Add some FODs to the positions, otherwise the method wont work
    extra = np.zeros((atoms.Ns, 3))
    atom_pyflosic = atoms.atom + ['X'] * len(extra)
    X_pyflosic = np.vstack((X, extra))

    # Do the pycom FOD generation
    atoms = Atoms(atom_pyflosic, X_pyflosic, elec_symbols=['X', None], spin=spin)
    p = parameters(mode='restricted')
    p.init_atoms(atoms)
    p.basis = basis
    p.pycom_loc = loc
    pc = pycom(mf=mf, p=p)
    pc.get_guess()

    # Get the actual FOD positions from the xyz file
    atom, X = read_xyz(f'{loc}_GUESS_COM.xyz')
    _, _, fods = split_atom_and_fod(atom, X)

    if clean:
        remove(p.log_name)
        remove(f'{loc}_GUESS_COM.xyz')
    return fods


def split_atom_and_fod(atom, X):
    '''Split atom and FOD coordinates.

    Args:
        atom : list
            Atom symbols.

        X : array
            Atom positions.

    Returns:
        Shortened atom types and coordinates, with FOD coordinates as a tuple(list, array, array).
    '''
    X_fod = []
    # Iterate in reverted order, because we may delete elements
    for ia in range(len(X) - 1, -1, -1):
        if atom[ia] == 'X':
            X_fod.append(X[ia])
            X = np.delete(X, ia, axis=0)
            del atom[ia]
    X_fod = np.asarray(X_fod)
    return atom, X, X_fod


def remove_core_fods(atoms, fods, method='auto', radius=0.1):
    '''Remove core FODs from a set of FOD coordinates.

    Args:
        atoms :
            Atoms object.

        fods : array
            FOD positions.

    Kwargs:
        method : str
            Method to remove FODs (case insensitive). 'Automatic' will remove the nearest FODs of
            atoms depending on the number of core states. 'Radius' removes FODs within a fixed
            radius around atoms, 'covalent' does the same, but scaled by the atoms covalent radius.

        radius : float
            Radius size for 'radius' method, covalent radius scaling factor for 'covalent'. No
            effect for the 'automatic' method.

    Returns:
        Valence FOD positions as an array.
    '''
    method = method.lower()
    if method in ('auto', 'automatic'):
        for ia in range(atoms.Natoms):
            n_core = symbol2number[atoms.atom[ia]] - atoms.Z[ia]
            # In the spin-paired case two electrons are one state
            # Since only core states are removed in pseudopotentials this value is divisible by 2
            # +1 to account for uneven amount of core fods (e.g., for hydrogen)
            n_core = (n_core + 1) // 2
            dist = norm(fods - atoms.X[ia], axis=1)
            idx = np.argsort(dist)
            # Remove core FODs with the smallest distance to the core
            fods = np.delete(fods, idx[:n_core], axis=0)
    else:
        for ia in range(atoms.Natoms):
            dist = norm(fods - atoms.X[ia], axis=1)
            if method in ('rad', 'radius'):
                # Remove FODs within a sphere of given radius around atoms
                fods = fods[dist > radius]
            elif method in ('cov', 'covalent'):
                # Remove FODs within a sphere scaled by the covalent radius around atoms
                fods = fods[dist > radius * covalent_radii[atoms.atom[ia]]]
    return fods
